# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Callable, Dict, List, Optional, Tuple, Union

from jax import numpy as jnp
from jaxtyping import Array, ArrayLike
from numpyro._typing import DistributionLike
from numpyro.distributions.distribution import enable_validation

import gwkokab
from gwkokab.inference import numpyro_poisson_likelihood, poisson_likelihood
from gwkokab.models import NPowerlawMGaussian
from gwkokab.models.npowerlawmgaussian._ncombination import (
    create_truncated_normal_distributions,
)
from gwkokab.models.utils import JointDistribution
from gwkokab.parameters import Parameters
from gwkokab.poisson_mean import PoissonMean
from kokab.utils.checks import check_min_concentration_for_beta_dist
from kokab.utils.common import expand_arguments
from kokab.utils.flowMC_based import flowMC_arg_parser, FlowMCBased
from kokab.utils.numpyro_based import numpyro_arg_parser, NumpyroBased
from kokab.utils.sage import Sage, sage_arg_parser


def where_fns_list(has_beta_spin: bool) -> Optional[List[Callable[..., Array]]]:
    where_fns = []

    if has_beta_spin:

        def mean_variance_check(N_pl: int, N_g: int, **kwargs) -> Array:
            if N_pl > 0:
                means_pl = jnp.stack(
                    [kwargs[f"chi{i}_mean_pl_{j}"] for j in range(N_pl) for i in (1, 2)]
                )
                vars_pl = jnp.stack(
                    [
                        kwargs[f"chi{i}_variance_pl_{j}"]
                        for j in range(N_pl)
                        for i in (1, 2)
                    ]
                )
            if N_g > 0:
                means_g = jnp.stack(
                    [kwargs[f"chi{i}_mean_g_{j}"] for j in range(N_g) for i in (1, 2)]
                )
                vars_g = jnp.stack(
                    [
                        kwargs[f"chi{i}_variance_g_{j}"]
                        for j in range(N_g)
                        for i in (1, 2)
                    ]
                )

            if N_pl > 0 and N_g > 0:
                means = jnp.concatenate([means_pl, means_g])
                variances = jnp.concatenate([vars_pl, vars_g])
            elif N_pl > 0:
                means = means_pl
                variances = vars_pl
            else:
                means = means_g
                variances = vars_g

            checks = check_min_concentration_for_beta_dist(means, variances)
            return jnp.all(checks)

        where_fns.append(mean_variance_check)
    return where_fns if len(where_fns) > 0 else None


class NPowerlawMGaussianCore(Sage):
    def __init__(
        self,
        N_pl: int,
        N_g: int,
        has_beta_spin: bool,
        has_truncated_normal_spin: bool,
        has_tilt: bool,
        has_eccentricity: bool,
        has_redshift: bool,
        has_cos_iota: bool,
        has_phi_12: bool,
        has_polarization_angle: bool,
        has_right_ascension: bool,
        has_sin_declination: bool,
        has_detection_time: bool,
        likelihood_fn: Callable[
            [
                Callable[..., DistributionLike],
                JointDistribution,
                Dict[str, DistributionLike],
                Dict[str, int],
                ArrayLike,
                PoissonMean,
                Optional[List[Callable[..., Array]]],
                Dict[str, Array],
            ],
            Callable,
        ],
        posterior_regex: str,
        posterior_columns: List[str],
        seed: int,
        prior_filename: str,
        selection_fn_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_buckets: int,
        threshold: float,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
    ) -> None:
        self.N_pl = N_pl
        self.N_g = N_g
        self.has_beta_spin = has_beta_spin
        self.has_truncated_normal_spin = has_truncated_normal_spin
        if self.has_truncated_normal_spin:
            gwkokab.models.npowerlawmgaussian._model.build_spin_distributions = (
                create_truncated_normal_distributions
            )
        self.has_tilt = has_tilt
        self.has_eccentricity = has_eccentricity
        self.has_redshift = has_redshift
        self.has_cos_iota = has_cos_iota
        self.has_phi_12 = has_phi_12
        self.has_polarization_angle = has_polarization_angle
        self.has_right_ascension = has_right_ascension
        self.has_sin_declination = has_sin_declination
        self.has_detection_time = has_detection_time

        super().__init__(
            likelihood_fn=likelihood_fn,
            model=NPowerlawMGaussian,
            posterior_regex=posterior_regex,
            posterior_columns=posterior_columns,
            seed=seed,
            prior_filename=prior_filename,
            selection_fn_filename=selection_fn_filename,
            poisson_mean_filename=poisson_mean_filename,
            sampler_settings_filename=sampler_settings_filename,
            analysis_name="n_pls_m_gs",
            n_buckets=n_buckets,
            threshold=threshold,
            debug_nans=debug_nans,
            profile_memory=profile_memory,
            check_leaks=check_leaks,
            where_fns=where_fns_list(has_beta_spin=has_beta_spin),
        )

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        return {
            "N_pl": self.N_pl,
            "N_g": self.N_g,
            "use_spin": self.has_beta_spin or self.has_truncated_normal_spin,
            "use_tilt": self.has_tilt,
            "use_eccentricity": self.has_eccentricity,
            "use_redshift": self.has_redshift,
            "use_cos_iota": self.has_cos_iota,
            "use_phi_12": self.has_phi_12,
            "use_polarization_angle": self.has_polarization_angle,
            "use_right_ascension": self.has_right_ascension,
            "use_sin_declination": self.has_sin_declination,
            "use_detection_time": self.has_detection_time,
        }

    @property
    def parameters(self) -> List[str]:
        names = [
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
        ]
        if self.has_beta_spin or self.has_truncated_normal_spin:
            names.append(Parameters.PRIMARY_SPIN_MAGNITUDE.value)
            names.append(Parameters.SECONDARY_SPIN_MAGNITUDE.value)
        if self.has_tilt:
            names.extend([Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value])
        if self.has_phi_12:
            names.append(Parameters.PHI_12.value)
        if self.has_eccentricity:
            names.append(Parameters.ECCENTRICITY.value)
        if self.has_redshift:
            names.append(Parameters.REDSHIFT.value)
        if self.has_right_ascension:
            names.append(Parameters.RIGHT_ASCENSION.value)
        if self.has_sin_declination:
            names.append(Parameters.SIN_DECLINATION.value)
        if self.has_detection_time:
            names.append(Parameters.DETECTION_TIME.value)
        if self.has_cos_iota:
            names.append(Parameters.COS_IOTA.value)
        if self.has_polarization_angle:
            names.append(Parameters.POLARIZATION_ANGLE.value)
        return names

    @property
    def model_parameters(self) -> List[str]:
        all_params: List[Tuple[str, int]] = [
            ("log_rate", self.N_pl + self.N_g),
            ("alpha_pl", self.N_pl),
            ("beta_pl", self.N_pl),
            ("m1_loc_g", self.N_g),
            ("m2_loc_g", self.N_g),
            ("m1_scale_g", self.N_g),
            ("m2_scale_g", self.N_g),
            ("m1_low_g", self.N_g),
            ("m2_low_g", self.N_g),
            ("m1_high_g", self.N_g),
            ("m2_high_g", self.N_g),
            ("mmax_pl", self.N_pl),
            ("mmin_pl", self.N_pl),
        ]

        if self.has_truncated_normal_spin:
            all_params.extend(
                [
                    ("chi1_high_g", self.N_g),
                    ("chi1_high_pl", self.N_pl),
                    ("chi1_loc_g", self.N_g),
                    ("chi1_loc_pl", self.N_pl),
                    ("chi1_low_g", self.N_g),
                    ("chi1_low_pl", self.N_pl),
                    ("chi1_scale_g", self.N_g),
                    ("chi1_scale_pl", self.N_pl),
                    ("chi2_high_g", self.N_g),
                    ("chi2_high_pl", self.N_pl),
                    ("chi2_loc_g", self.N_g),
                    ("chi2_loc_pl", self.N_pl),
                    ("chi2_low_g", self.N_g),
                    ("chi2_low_pl", self.N_pl),
                    ("chi2_scale_g", self.N_g),
                    ("chi2_scale_pl", self.N_pl),
                ]
            )
        if self.has_beta_spin:
            all_params.extend(
                [
                    ("chi1_mean_g", self.N_g),
                    ("chi1_mean_pl", self.N_pl),
                    ("chi1_variance_g", self.N_g),
                    ("chi1_variance_pl", self.N_pl),
                    ("chi2_mean_g", self.N_g),
                    ("chi2_mean_pl", self.N_pl),
                    ("chi2_variance_g", self.N_g),
                    ("chi2_variance_pl", self.N_pl),
                ]
            )

        if self.has_tilt:
            all_params.extend(
                [
                    ("cos_tilt_zeta_g", self.N_g),
                    ("cos_tilt_zeta_pl", self.N_pl),
                    ("cos_tilt1_scale_g", self.N_g),
                    ("cos_tilt1_scale_pl", self.N_pl),
                    ("cos_tilt2_scale_g", self.N_g),
                    ("cos_tilt2_scale_pl", self.N_pl),
                ]
            )

        if self.has_phi_12:
            all_params.extend(
                [
                    (Parameters.PHI_12.value + "_high_g", self.N_g),
                    (Parameters.PHI_12.value + "_high_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_loc_g", self.N_g),
                    (Parameters.PHI_12.value + "_loc_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_low_g", self.N_g),
                    (Parameters.PHI_12.value + "_low_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_scale_g", self.N_g),
                    (Parameters.PHI_12.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_eccentricity:
            all_params.extend(
                [
                    ("ecc_high_g", self.N_g),
                    ("ecc_high_pl", self.N_pl),
                    ("ecc_loc_g", self.N_g),
                    ("ecc_loc_pl", self.N_pl),
                    ("ecc_low_g", self.N_g),
                    ("ecc_low_pl", self.N_pl),
                    ("ecc_scale_g", self.N_g),
                    ("ecc_scale_pl", self.N_pl),
                ]
            )

        if self.has_redshift:
            all_params.extend(
                [
                    (Parameters.REDSHIFT.value + "_kappa_g", self.N_g),
                    (Parameters.REDSHIFT.value + "_kappa_pl", self.N_pl),
                    (Parameters.REDSHIFT.value + "_z_max_g", self.N_g),
                    (Parameters.REDSHIFT.value + "_z_max_pl", self.N_pl),
                ]
            )

        if self.has_right_ascension:
            all_params.extend(
                [
                    (Parameters.RIGHT_ASCENSION.value + "_high_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_high_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_loc_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_loc_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_low_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_low_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_scale_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_sin_declination:
            all_params.extend(
                [
                    (Parameters.SIN_DECLINATION.value + "_high_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_high_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_loc_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_loc_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_low_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_low_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_scale_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_detection_time:
            all_params.extend(
                [
                    (Parameters.DETECTION_TIME.value + "_high_g", self.N_g),
                    (Parameters.DETECTION_TIME.value + "_high_pl", self.N_pl),
                    (Parameters.DETECTION_TIME.value + "_low_g", self.N_g),
                    (Parameters.DETECTION_TIME.value + "_low_pl", self.N_pl),
                ]
            )

        if self.has_cos_iota:
            all_params.extend(
                [
                    (Parameters.COS_IOTA.value + "_high_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_high_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_loc_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_loc_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_low_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_low_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_scale_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_polarization_angle:
            all_params.extend(
                [
                    (Parameters.POLARIZATION_ANGLE.value + "_high_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_high_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_loc_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_loc_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_low_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_low_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_scale_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_scale_pl", self.N_pl),
                ]
            )

        extended_params = []
        for params in all_params:
            extended_params.extend(expand_arguments(*params))
        return extended_params


class NPowerlawMGaussianFSage(NPowerlawMGaussianCore, FlowMCBased):
    pass


class NPowerlawMGaussianNSage(NPowerlawMGaussianCore, NumpyroBased):
    pass


def model_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--n-pl",
        type=int,
        help="Number of power-law components in the mass model.",
    )
    model_group.add_argument(
        "--n-g",
        type=int,
        help="Number of Gaussian components in the mass model.",
    )

    spin_group = model_group.add_mutually_exclusive_group()
    spin_group.add_argument(
        "--add-beta-spin",
        action="store_true",
        help="Include beta spin parameters in the model.",
    )
    spin_group.add_argument(
        "--add-truncated-normal-spin",
        action="store_true",
        help="Include truncated normal spin parameters in the model.",
    )

    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    model_group.add_argument(
        "--add-truncated-normal-eccentricity",
        action="store_true",
        help="Include truncated normal eccentricity in the model.",
    )
    model_group.add_argument(
        "--add-cos-iota",
        action="store_true",
        help="Include cos_iota parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-12",
        action="store_true",
        help="Include phi_12 parameter in the model",
    )
    model_group.add_argument(
        "--add-polarization-angle",
        action="store_true",
        help="Include polarization_angle parameter in the model",
    )
    model_group.add_argument(
        "--add-right-ascension",
        action="store_true",
        help="Include right_ascension parameter in the model",
    )
    model_group.add_argument(
        "--add-sin-declination",
        action="store_true",
        help="Include sin_declination parameter in the model",
    )
    model_group.add_argument(
        "--add-detection-time",
        action="store_true",
        help="Include detection_time parameter in the model",
    )
    return parser


def f_main() -> None:
    enable_validation()

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = flowMC_arg_parser(parser)

    args = parser.parse_args()

    NPowerlawMGaussianFSage(
        N_pl=args.n_pl,
        N_g=args.n_g,
        has_beta_spin=args.add_beta_spin,
        has_truncated_normal_spin=args.add_truncated_normal_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_truncated_normal_eccentricity,
        has_redshift=args.add_redshift,
        has_cos_iota=args.add_cos_iota,
        has_phi_12=args.add_phi_12,
        has_polarization_angle=args.add_polarization_angle,
        has_right_ascension=args.add_right_ascension,
        has_sin_declination=args.add_sin_declination,
        has_detection_time=args.add_detection_time,
        likelihood_fn=poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()


def n_main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = numpyro_arg_parser(parser)

    args = parser.parse_args()

    NPowerlawMGaussianNSage(
        N_pl=args.n_pl,
        N_g=args.n_g,
        has_beta_spin=args.add_beta_spin,
        has_truncated_normal_spin=args.add_truncated_normal_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_truncated_normal_eccentricity,
        has_redshift=args.add_redshift,
        has_cos_iota=args.add_cos_iota,
        has_phi_12=args.add_phi_12,
        has_polarization_angle=args.add_polarization_angle,
        has_right_ascension=args.add_right_ascension,
        has_sin_declination=args.add_sin_declination,
        has_detection_time=args.add_detection_time,
        likelihood_fn=numpyro_poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import warnings
from collections.abc import Sequence
from typing import List, Optional, Tuple

import equinox as eqx
import h5py
import jax
import numpy as np
import pandas as pd
from jax import nn as jnn, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray, PyTree
from numpyro.util import is_prng_key


__all__ = [
    "load_model",
    "make_model",
    "mse_loss_fn",
    "predict",
    "read_data",
    "save_model",
]


@eqx.filter_value_and_grad
@eqx.filter_jit
def mse_loss_fn(
    model: PyTree, x: Array, y: Array, batch_size: Optional[int] = 256
) -> Array:
    """Mean squared error loss function.

    Parameters
    ----------
    model : PyTree
        Model to approximate the log of the VT function
    x : Array
        input data
    y : Array
        output data
    batch_size : Optional[int], optional
        batch size for training, by default 256. This is used to avoid OOM errors when
        training large datasets. If None, the entire dataset will be trained
        sequentially. This is not recommended for large datasets.

    Returns
    -------
    Array
        mean squared error
    """
    y_pred = jax.lax.map(model, x, batch_size=batch_size)
    return jnp.mean(jnp.square(y - y_pred))  # mean squared error


@eqx.filter_jit
def predict(model: PyTree, x: Array, batch_size: Optional[int] = 256) -> Array:
    """Predict the output of the model given the input data.

    Parameters
    ----------
    model : PyTree
        Model to approximate the log of the VT function
    x : Array
        input data
    batch_size : Optional[int], optional
        batch size for prediction, by default 256. This is used to avoid OOM errors when
        predicting large datasets. If None, the entire dataset will be predicted
        sequentially. This is not recommended for large datasets.

    Returns
    -------
    Array
        predicted output
    """
    return jax.lax.map(model, x, batch_size=batch_size)


def read_data(data_path: str, keys: Sequence[str]) -> pd.DataFrame:
    """Read the data from the given path.

    Parameters
    ----------
    data_path : str
        path to the data
    keys : Sequence[str]
        keys to read from the data file

    Returns
    -------
    pd.DataFrame
        data in a DataFrame
    """
    with h5py.File(data_path, "r") as vt_file:
        df = pd.DataFrame(data={key: np.array(vt_file[key]).flatten() for key in keys})
    return df


def make_model(
    *,
    key: PRNGKeyArray,
    input_layer: int,
    output_layer: int,
    width_size: int,
    depth: int,
) -> eqx.nn.MLP:
    """Make a neural network model to approximate the log of the VT function.

    Parameters
    ----------
    key : PRNGKeyArray
        jax random key
    input_layer : int
        input layer of the model
    output_layer : int
        output layer of the model
    width_size : int
        width size of the model
    depth : int
        depth of the model

    Returns
    -------
    eqx.nn.MLP
        neural network model
    """
    assert is_prng_key(key)

    model = eqx.nn.MLP(
        in_size=input_layer,
        out_size=output_layer,
        width_size=width_size,
        depth=depth,
        activation=jnn.relu,
        key=key,
    )
    return model


def save_model(
    *,
    filepath: str,
    datafilepath: str,
    model: eqx.nn.MLP,
    names: Optional[Sequence[str]] = None,
    is_log: bool = False,
) -> None:
    """Save the model to the given file.

    Parameters
    ----------
    filepath : str
        Name of the file to save the model
    datafilepath : str
        Path to the data file, used to save the names of the parameters
    model : eqx.nn.MLP
        Model to approximate the log of the VT function
    names : Optional[Sequence[str]], optional
        names of the parameters, by default None
    is_log : bool, optional
        Whether the model was trained on log-transformed data, by default False
    """
    if not filepath.endswith(".hdf5"):
        if "." in filepath:
            old_filename = filepath
            filepath = filepath.split(".")[0] + ".hdf5"
            warnings.warn(
                f"Neural VT path does not end with .hdf5: {old_filename}. Saving to {filepath} instead."
            )

    with h5py.File(datafilepath, "r") as f:
        # read all attributes from the data file
        attr = dict(f.attrs)
    model: eqx.nn.MLP = model._fun  # type: ignore
    with h5py.File(filepath, "w") as f:
        for key, value in attr.items():
            f.attrs[key] = value  # copy attributes from the data file
        if names is not None:
            f.create_dataset("names", data=np.array(names, dtype="S"))
        f.create_dataset("in_size", data=model.in_size)  # type: ignore
        f.create_dataset("out_size", data=model.out_size)  # type: ignore
        f.create_dataset("width_size", data=model.width_size)  # type: ignore
        f.create_dataset("depth", data=model.depth)  # type: ignore
        f.attrs["is_log"] = is_log
        num_layers = len(model.layers)  # type: ignore
        for i in range(num_layers):
            layer_i = f.create_group(f"layer_{i}")
            layer_i.create_dataset(f"weight_{i}", data=model.layers[i].weight)  # type: ignore
            layer_i.create_dataset(f"bias_{i}", data=model.layers[i].bias)  # type: ignore


def load_model(filename: str) -> Tuple[List[str], eqx.nn.MLP]:
    """Load the model from the given file.

    Parameters
    ----------
    filename : str
        Name of the file to load the model

    Returns
    -------
    Tuple[List[str], eqx.nn.MLP]
        names of the parameters and the model
    """
    with h5py.File(filename, "r") as f:
        names = f["names"][:]
        names = names.astype(str).tolist()
        in_size = int(f["in_size"][()])
        out_size = int(f["out_size"][()])
        width_size = int(f["width_size"][()])
        depth = int(f["depth"][()])
        new_model = eqx.nn.MLP(
            in_size=in_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            activation=jnn.relu,
            key=jrd.PRNGKey(0),
        )
        i = 0
        while f.get(f"layer_{i}"):
            layer_i = f[f"layer_{i}"]
            weight_i = jax.device_put(np.asarray(layer_i[f"weight_{i}"][:]))
            bias_i = jax.device_put(np.asarray(layer_i[f"bias_{i}"][:]))
            new_model = eqx.tree_at(lambda m: m.layers[i].weight, new_model, weight_i)
            new_model = eqx.tree_at(lambda m: m.layers[i].bias, new_model, bias_i)
            i += 1

    return names, new_model



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import importlib.metadata
from typing import Final


__version__: Final[str] = importlib.metadata.version("gwkokab")


from . import (
    constants as constants,
    cosmology as cosmology,
    errors as errors,
    inference as inference,
    models as models,
    parameters as parameters,
    poisson_mean as poisson_mean,
    population as population,
    utils as utils,
    vts as vts,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

from jax import numpy as jnp
from jaxtyping import Array
from numpyro.distributions import TruncatedNormal

from gwkokab.models import Wysocki2019MassModel
from gwkokab.models.utils import JointDistribution, ScaledMixture


def EccentricityMattersModel(
    log_rate: Array,
    alpha_m: Array,
    mmin: Array,
    mmax: Array,
    loc: Array,
    scale: Array,
    low: Array,
    high: Array,
    *,
    validate_args: Optional[bool] = None,
) -> ScaledMixture:
    comp_dist = JointDistribution(
        Wysocki2019MassModel(
            alpha_m=alpha_m, mmin=mmin, mmax=mmax, validate_args=validate_args
        ),
        TruncatedNormal(
            loc=loc, scale=scale, low=low, high=high, validate_args=validate_args
        ),
    )
    return ScaledMixture(
        log_scales=jnp.array([log_rate]),
        component_distributions=[comp_dist],
        support=comp_dist.support,
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentParser
from collections.abc import Callable
from glob import glob
from typing import Dict, List, Optional, Tuple, Union

import jax
import numpy as np
from jaxtyping import Array, ArrayLike
from loguru import logger
from numpyro._typing import DistributionLike
from numpyro.distributions import Distribution

from gwkokab.models.utils import JointDistribution
from gwkokab.poisson_mean import PoissonMean
from gwkokab.utils.tools import warn_if
from kokab.utils.common import get_posterior_data
from kokab.utils.literals import LOG_REF_PRIOR_NAME
from kokab.utils.poisson_mean_parser import read_pmean

from .guru import Guru
from .jenks import pad_and_stack


class Sage(Guru):
    def __init__(
        self,
        likelihood_fn: Callable[
            [
                Callable[..., DistributionLike],
                JointDistribution,
                Dict[str, DistributionLike],
                Dict[str, int],
                ArrayLike,
                PoissonMean,
                Optional[List[Callable[..., Array]]],
                Dict[str, Array],
            ],
            Callable,
        ],
        model: Union[Distribution, Callable[..., Distribution]],
        where_fns: Optional[List[Callable[..., Array]]],
        posterior_regex: str,
        posterior_columns: List[str],
        seed: int,
        prior_filename: str,
        selection_fn_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_buckets: int,
        threshold: float,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
        analysis_name: str = "",
    ) -> None:
        """
        Parameters
        ----------
        model : Union[Distribution, Callable[..., Distribution]]
            model to be used in the Sage class. It can be a Distribution or a callable
            that returns a Distribution.
        where_fns : Optional[List[Callable[..., Array]]]
            List of functions to apply to the data before passing it to the model.
        posterior_regex : str
            Regular expression to match posterior files.
        posterior_columns : List[str]
            List of columns to extract from the posterior files.
        seed : int
            Seed for the random number generator.
        prior_filename : str
            Path to the JSON file containing the prior distributions.
        selection_fn_filename : str
            Path to the JSON file containing the selection function.
        poisson_mean_filename : str
            Path to the JSON file containing the Poisson mean configuration.
        sampler_settings_filename : str
            Path to the JSON file containing the sampler settings.
        n_buckets : int
            Number of buckets to use for padding and stacking the data.
        threshold : float
            Threshold for padding and stacking the data.
        debug_nans : bool, optional
            If True, checks for NaNs in each computation. See details in the
            [documentation](https://jax.readthedocs.io/en/latest/_autosummary/jax.debug_nans.html#jax.debug_nans),
            by default False
        profile_memory : bool, optional
            If True, enables memory profiling, by default False
        check_leaks : bool, optional
            If True, checks for JAX Tracer leaks. See details in the
            [documentation](https://jax.readthedocs.io/en/latest/_autosummary/jax.checking_leaks.html#jax.checking_leaks),
            by default False
        analysis_name : str, optional
            Name of the analysis, by default ""
        """
        assert all(letter.isalpha() or letter == "_" for letter in analysis_name), (
            "Analysis name must contain only letters and underscores."
        )
        self.likelihood_fn = likelihood_fn
        self.n_buckets = n_buckets
        self.posterior_columns = posterior_columns
        self.posterior_regex = posterior_regex
        self.threshold = threshold
        self.where_fns = where_fns
        self.set_rng_key(seed=seed)

        super().__init__(
            analysis_name=analysis_name,
            check_leaks=check_leaks,
            debug_nans=debug_nans,
            model=model,
            poisson_mean_filename=poisson_mean_filename,
            prior_filename=prior_filename,
            profile_memory=profile_memory,
            sampler_settings_filename=sampler_settings_filename,
            selection_fn_filename=selection_fn_filename,
        )

    def read_data(
        self,
    ) -> Tuple[int, float, Tuple[Array, ...], Tuple[Array, ...], Tuple[Array, ...]]:
        data = get_posterior_data(glob(self.posterior_regex), self.posterior_columns)
        if LOG_REF_PRIOR_NAME in self.posterior_columns:
            idx = self.posterior_columns.index(LOG_REF_PRIOR_NAME)
            self.posterior_columns.pop(idx)
            warn_if(
                True,
                msg="Please ensure that reference prior is in the last column of the posteriors.",
            )
            log_ref_priors = [d[..., idx] for d in data]
            data = [np.delete(d, idx, axis=-1) for d in data]
        else:
            log_ref_priors = [np.zeros(d.shape[:-1]) for d in data]
        assert len(data) == len(log_ref_priors), (
            "Data and log reference priors must have the same length"
        )
        _data_group, _log_ref_priors_group, _masks_group = pad_and_stack(
            data, log_ref_priors, n_buckets=self.n_buckets, threshold=self.threshold
        )

        _data_group = tuple(_data_group)
        _log_ref_priors_group = tuple(_log_ref_priors_group)
        _masks_group = tuple(_masks_group)

        data_group: Tuple[Array] = jax.block_until_ready(
            jax.device_put(_data_group, may_alias=True)
        )
        log_ref_priors_group: Tuple[Array] = jax.block_until_ready(
            jax.device_put(_log_ref_priors_group, may_alias=True)
        )
        masks_group: Tuple[Array] = jax.block_until_ready(
            jax.device_put(_masks_group, may_alias=True)
        )

        logger.debug(
            "data_group.shape: {shape}",
            shape=", ".join([str(d.shape) for d in data_group]),
        )
        logger.debug(
            "log_ref_priors_group.shape: {shape}",
            shape=", ".join([str(d.shape) for d in log_ref_priors_group]),
        )
        logger.debug(
            "masks_group.shape: {shape}",
            shape=", ".join([str(d.shape) for d in masks_group]),
        )

        n_events = len(data)
        sum_log_size = sum([np.log(d.shape[0]) for d in data])
        log_constants = -sum_log_size  # -Σ log(M_i)

        return n_events, log_constants, data_group, log_ref_priors_group, masks_group

    def run(self) -> None:
        constants, dist_fn, priors, variables, variables_index = self.bake_model()

        n_events, log_constants, data_group, log_ref_priors_group, masks_group = (
            self.read_data()
        )

        ERate_obj = read_pmean(
            self.rng_key,
            self.parameters,
            self.poisson_mean_filename,
            self.selection_fn_filename,
        )

        log_constants += n_events * np.log(ERate_obj.time_scale)  # type: ignore

        logpdf = self.likelihood_fn(
            dist_fn=dist_fn,
            priors=priors,
            variables=variables,
            variables_index=variables_index,
            log_constants=log_constants,
            ERate_obj=ERate_obj,
            where_fns=self.where_fns,
            constants=constants,  # type: ignore
        )

        self.driver(
            logpdf=logpdf,
            priors=priors,
            data={
                "data_group": data_group,
                "log_ref_priors_group": log_ref_priors_group,
                "masks_group": masks_group,
            },
            labels=sorted(variables.keys()),
        )


def sage_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Sage script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    sage_group = parser.add_argument_group("Sage Options")
    sage_group.add_argument(
        "--posterior-regex",
        help="Regex for the posterior samples.",
        type=str,
        required=True,
    )
    sage_group.add_argument(
        "--posterior-columns",
        help="Columns of the posterior samples.",
        nargs="+",
        type=str,
        required=True,
    )

    optm_group = parser.add_argument_group("Optimization Options")
    optm_group.add_argument(
        "--n-buckets",
        help="Number of buckets for the data arrays to be split into. "
        "This is useful for large datasets to avoid memory issues. "
        "See https://github.com/gwkokab/gwkokab/issues/568 for more details.",
        type=int,
        default=None,
    )
    optm_group.add_argument(
        "--threshold",
        help="Threshold to determine best number of buckets, if the number of buckets "
        "is not specified. It should be between 0 and 100.",
        type=float,
        default=3.0,
    )

    return parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Bucketing and padding utilities using Jenks' Natural Breaks algorithm.

This module provides functionality to partition sequences of arrays into buckets based
on their sizes using Jenks' Natural Breaks algorithm, then pad and stack these arrays
within each bucket for uniform shape. This is useful for managing memory usage when
processing large datasets.
"""

from collections.abc import Sequence
from typing import Optional, TypeVar, Union

import jenkspy
import numpy as np
from jax import numpy as jnp
from jaxtyping import Array
from loguru import logger

from gwkokab.utils.tools import error_if


__all__ = ["pad_and_stack"]

T = TypeVar("T", bound=Array | np.ndarray | int)


def _loss(bucket: Sequence[int]) -> float:
    r"""Calculate the loss for a given bucket of data.

    .. math::
        \text{loss} = \frac{n \cdot \max(\mathrm{bucket}) - \text{sum}(\mathrm{bucket})}{n \cdot \max(\mathrm{bucket})} \times 100

    Parameters
    ----------
    bucket : Sequence[int]
        A sequence of integers representing a bucket of data.

    Returns
    -------
    float
        The loss percentage for the bucket, calculated as described above.
    """

    total = sum(bucket)
    return (len(bucket) * max(bucket) - total) / (len(bucket) * max(bucket)) * 100.0


def _total_loss(subsets: Sequence[Sequence[int]]) -> float:
    total_loss = (
        (
            sum(len(subset) * max(subset) for subset in subsets)
            - sum(sum(subset) for subset in subsets)
        )
        / sum(len(subset) * max(subset) for subset in subsets)
        * 100.0
    )
    return total_loss


def _right_most_index(data: Sequence[int], value: int) -> int:
    """Using Binary Search to find the rightmost index of a value in a sorted list.

    Parameters
    ----------
    data : Sequence[int]
        Sorted list of integers.
    value : int
        The value to find the rightmost index of.

    Returns
    -------
    int
        The rightmost index of the value in the list, or -1 if not found.

    Raises
    ------
    ValueError
        If the value is not found in the list or if the right index is out of bounds.
    """
    low, high = 0, len(data) - 1
    result = -1
    while low <= high:
        mid = (low + high) >> 1
        if data[mid] == value:
            result = mid
            low = mid + 1
        elif value < data[mid]:
            high = mid - 1
        else:
            low = mid + 1
    error_if(result == -1, msg=f"Value {value} not found in the list.")
    return result


def _jenks_natural_breaks(
    data: Sequence[int], n_buckets: int, verbose: bool = True
) -> Sequence[int]:
    """Calculate Jenks' Natural Breaks for a given data set and number of buckets.

    Parameters
    ----------
    data : Sequence[int]
        A sequence of integers representing the data to be bucketed.
    n_buckets : int
        The number of buckets to create from the data.
    verbose : bool, optional
        If True, logs the total loss and losses for each bucket. Default is True.

    Returns
    -------
    Sequence[int]
        A sequence of indices representing the rightmost index of each bucket's break
        value in the sorted data.
    """
    error_if(
        n_buckets < 1 or n_buckets > len(data),
        msg=f"Number of buckets {n_buckets} must be between 1 and {len(data)}",
    )
    if n_buckets == 1:
        return [0, len(data) - 1]
    breaks = jenkspy.jenks_breaks(data, n_classes=n_buckets)
    n_unique_breaks = len(set(breaks))
    if n_unique_breaks != n_buckets + 1:
        if n_unique_breaks != n_buckets + 1:
            msg = f"{n_buckets} buckets requested, but less than {n_buckets + 1} unique breaks found. "
            msg += "Consider halving the number of buckets or using a different method."
            if verbose:
                logger.error(msg)
            raise ValueError(msg)
    indexes = [_right_most_index(data, break_value) for break_value in breaks]
    return indexes


def _get_subsets(data: Sequence[T], indexes: Sequence[int]) -> Sequence[Sequence[T]]:
    """Get the subset of data based on the provided indexes.

    Parameters
    ----------
    data : Sequence[T]
        A sequence of data from which subsets will be created.
    indexes : Sequence[int]
        A sequence of indices that define the boundaries of the subsets.

    Returns
    -------
    Sequence[Sequence[T]]
        A sequence of subsets of the data, where each subset corresponds to the
        indices defined in `indexes`.
    """
    return [data[indexes[i] : indexes[i + 1]] for i in range(len(indexes) - 2)] + [
        data[indexes[-2] : indexes[-1] + 1]
    ]


def _partition_data_for_bucket(
    data: Sequence[int], n_buckets: int, verbose: bool = True
) -> tuple[Sequence[int], float]:
    """Partition the data into buckets using Jenks' Natural Breaks.

    Parameters
    ----------
    data : Sequence[int]
        A sequence of integers representing the data to be partitioned.
    n_buckets : int
        The number of buckets to create from the data.
    verbose : bool, optional
        If True, logs the total loss and losses for each bucket, by default True

    Returns
    -------
    tuple[Sequence[int], float]
        A tuple containing:
        - A sequence of indices representing the rightmost index of each bucket's break
          value in the sorted data.
        - The total loss percentage for the partitioned data.
    """
    indexes = _jenks_natural_breaks(data, n_buckets, verbose=verbose)
    subsets = _get_subsets(data, indexes)
    total_loss = _total_loss(subsets)
    if verbose:
        logger.info("Total loss of all buckets: {:.4f}%", _total_loss(subsets))

        losses = [_loss(subset) for subset in subsets]
        logger.info(
            "Losses for each bucket: {}", ", ".join(f"{loss:.4f}%" for loss in losses)
        )

    return indexes, total_loss


def _partition_data(
    data: Sequence[int], n_buckets: Optional[int] = None, threshold: float = 3.0
) -> Sequence[int]:
    if len(set(data)) == 1:
        logger.debug(
            "All elements in the data are identical in size. Returning single bucket."
        )
        return [0, len(data) - 1]
    if n_buckets is not None:
        indexes, total_loss = _partition_data_for_bucket(data, n_buckets)
        return indexes

    total_losses = []
    n_buckets_count = 1
    while True:
        try:
            _, total_loss = _partition_data_for_bucket(
                data, n_buckets_count, verbose=False
            )
        except ValueError:
            break
        total_losses.append(total_loss)
        n_buckets_count += 1

    d1 = np.diff(total_losses)
    loss_reduction = -d1

    below_thresh_idx = np.where(loss_reduction < threshold)[0]
    error_if(
        len(below_thresh_idx) == 0,
        msg="No suitable number of buckets found with the given threshold. "
        "Consider increasing the threshold.",
    )

    n_buckets = int(below_thresh_idx[0] + 1)  # +1 because diff shifts index
    logger.info(
        f"By first derivative of total loss with {threshold}% threshold, "
        f"recommended number of buckets is {n_buckets}."
    )
    indexes, _ = _partition_data_for_bucket(data, n_buckets, verbose=True)

    return indexes


def _pad_and_stack_bucket(bucket: Sequence[Union[Array, np.ndarray]]) -> Array:
    """Pad a bucket of arrays to the size of the largest array in the bucket and stack
    them.

    Parameters
    ----------
    bucket : Sequence[Union[Array, np.ndarray]]
        A sequence of arrays to pad and stack.

    Returns
    -------
    Array
        A stacked array where each array in the bucket is padded to the size of the
        largest array in the bucket. The shape of the returned array is (n_buckets, max_size,
        ...), where n_buckets is the number of arrays in the bucket and max_size is the size
        of the largest array in the bucket.
    """
    max_size = max(arr.shape[0] for arr in bucket)
    bucket_padded = [
        jnp.pad(b, ((0, max_size - b.shape[0]),) + ((0, 0),) * (b.ndim - 1))
        for b in bucket
    ]
    return jnp.stack(bucket_padded, axis=0)


def _bucket_mask(bucket: Sequence[Union[Array, np.ndarray]]) -> Array:
    """Create a mask for a bucket of arrays.

    Parameters
    ----------
    bucket : Sequence[Union[Array, np.ndarray]]
        A sequence of arrays to create a mask for.

    Returns
    -------
    Array
        A boolean array where True indicates a valid element and False indicates a padded
        element. The shape of the returned array is (n_buckets, max_size), where
        n_buckets is the number of arrays in the bucket and max_size is the size of the
        largest array in the bucket.
    """
    max_size = max(arr.shape[0] for arr in bucket)
    masks = [
        jnp.pad(
            jnp.ones(arr.shape[0], dtype=bool),
            (0, max_size - arr.shape[0]),
            constant_values=False,
        )
        for arr in bucket
    ]
    return jnp.stack(masks, axis=0)


def pad_and_stack(
    *arrays: Sequence[Union[Array, np.ndarray]],
    n_buckets: Optional[int],
    threshold: float,
) -> Sequence[Sequence[Array]]:
    """Pad and stack multiple arrays into buckets.

    Parameters
    ----------
    *arrays : Sequence[Union[Array, np.ndarray]]
        Variable number of array sequences to be bucketed. All sequences must
        have the same length.
    n_buckets : Optional[int]
        The number of buckets to create from the data. If None, the function will
        partition the data into buckets based on the sizes of the arrays.
    threshold : float
        if :code:`n_buckets` is None, this value is used to determine the maximum size of
        the buckets.

    Returns
    -------
    Sequence[Sequence[Array]]
        A sequence of padded arrays, where each array corresponds to a bucket.
        The last element of the sequence is a mask indicating which elements are valid
        and which are padded.
    """
    error_if(
        not (0.0 <= threshold <= 100.0),
        msg="Threshold must be between 0 and 100.",
    )
    error_if(
        not all(len(arrays[0]) == len(arr) for arr in arrays),
        msg="All arrays must have the same length.",
    )
    index_and_size = [(i, arr.size) for i, arr in enumerate(arrays[0])]
    index_and_size = sorted(index_and_size, key=lambda x: x[1])
    sizes = [size for _, size in index_and_size]
    shuffle_indices = [i for i, _ in index_and_size]
    shuffled_arrays = [[arrays_j[i] for i in shuffle_indices] for arrays_j in arrays]

    subset_indices = _partition_data(sizes, n_buckets, threshold)

    subsets_arrays = [_get_subsets(sa, subset_indices) for sa in shuffled_arrays]

    padded_subsets = [
        [_pad_and_stack_bucket(s) for s in subset] for subset in subsets_arrays
    ]
    masks = [_bucket_mask(s) for s in subsets_arrays[0]]

    return padded_subsets + [masks]



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from functools import partial
from typing import Any, Callable, Dict, Tuple

from jax import lax, random as jrd
from numpyro.distributions.distribution import Distribution


class Bake(object):
    def __init__(self, dist: Distribution | Callable[..., Distribution]) -> None:
        """It is designed to be a simple and flexible way to define a distribution for
        the inference. It has a similar interface to the
        :class:`~numpyro.distributions.distribution.Distribution` class, but it allows
        for the parameters of the distribution to be fixed or variable.

        Parameters
        ----------
        dist : Distribution | Callable[..., Distribution]
            A distribution or a function that returns a distribution
        """
        self._dist = dist

    def __call__(self, **kwargs: Any) -> "Bake":
        """Set the parameters of the distribution.

        Returns
        -------
        Self
            The Bake object

        Raises
        ------
        ValueError
            If the type of a parameter is invalid
        """
        constants: Dict[str, int | float | None] = dict()
        variables: Dict[str, Distribution] = dict()
        duplicates: Dict[str, str] = dict()
        for key, value in kwargs.items():
            if value is None:
                constants[key] = None
            elif isinstance(value, Distribution):
                variables[key] = value
            elif isinstance(value, (int, float)):
                constants[key] = lax.stop_gradient(value)
            elif isinstance(value, str):
                continue
            else:
                raise ValueError(
                    f"Parameter {key} has an invalid type {type(value)}: {value}"
                )
        for key, value in kwargs.items():
            if isinstance(value, str):
                if value in constants:
                    constants[key] = constants[value]
                elif value in variables:
                    duplicates[key] = value
        self.constants = constants
        self.variables = variables
        self.duplicates = duplicates
        return self

    def get_dist(
        self,
    ) -> Tuple[
        Dict[str, int | float | None],
        Dict[str, Distribution],
        Dict[str, str],
        Callable[..., Distribution],
    ]:
        """Return the distribution with the fixed parameters set.

        Returns
        -------
        Tuple[Dict[str, int | float | None], Dict[str, Distribution], Dict[str, str], Callable[..., Distribution]]
            A tuple containing the distribution with the fixed parameters set
            and a function that returns the distribution with the fixed parameters
            set.
        """
        return (
            self.constants,
            self.variables,
            self.duplicates,
            partial(self._dist, **self.constants),
        )

    def get_dummy(self) -> Distribution:
        """Return a dummy distribution for debug and testing purposes.

        Returns
        -------
        Distribution
            A dummy distribution
        """
        key = jrd.PRNGKey(0)
        variables = {
            name: prior.sample(key=key, sample_shape=())
            for name, prior in self.variables.items()
        }
        duplicates = {name: variables[value] for name, value in self.duplicates.items()}
        return self._dist(**self.constants, **variables, **duplicates)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from typing import Any, Dict, List, Literal, Union

from jaxtyping import PRNGKeyArray
from numpyro._typing import DistributionLike

from gwkokab.models.utils import JointDistribution
from gwkokab.poisson_mean import PoissonMean
from gwkokab.utils.tools import error_if

from .common import get_dist, vt_json_read_and_process


ProposalDistArgType = Dict[str, Union[Literal["self"], List[Dict[str, Any]]]]
PoissonMeanConfig = Dict[
    str,
    Union[
        int,  # sample size
        float,  # scale
        ProposalDistArgType,  # proposal_dists
    ],
]


def parse_distribution(
    per_component_dist: Union[str, List[Dict[str, Any]]],
) -> Union[Literal["self"], DistributionLike]:
    """Parse the proposal distribution.

    Parameters
    ----------
    per_component_dist : Union[str, List[Dict[str, Any]]]
        The proposal distribution

    Returns
    -------
    Union[Literal[&quot;self&quot;], DistributionLike]
        The proposal distribution

    Raises
    ------
    ValueError
        If the distribution format is invalid.
    """
    if isinstance(per_component_dist, str):
        error_if(
            per_component_dist.strip().lower() != "self",
            msg="The key in the proposal distribution must be 'self'.",
        )
        return "self"
    elif isinstance(per_component_dist, list):
        error_if(
            not all(isinstance(dist_dict, dict) for dist_dict in per_component_dist),
            msg="The proposal distribution must be a list of dictionaries.",
        )
        return JointDistribution(
            *[get_dist(dist_dict) for dist_dict in per_component_dist],
            validate_args=True,
        )
    raise ValueError("Invalid distribution format")


def poisson_mean_parser(filepath: str) -> PoissonMeanConfig:
    """Parse the JSON file containing the configuration of the Poisson mean.

    Parameters
    ----------
    filepath : str
        The path to the JSON file containing the configuration of the Poisson mean.

    Returns
    -------
    PoissonMeanConfig
        The configuration of the Poisson mean.
    """
    with open(filepath, "r") as file:
        pmean_json: PoissonMeanConfig = json.load(file)
    try:
        proposal_distribution_dict: ProposalDistArgType = pmean_json.pop(
            "proposal_dists"
        )
        proposal_distribution = [
            parse_distribution(dist) for dist in proposal_distribution_dict
        ]
        pmean_json["proposal_dists"] = proposal_distribution
    except KeyError:
        pass
    return pmean_json


def read_pmean(
    key: PRNGKeyArray, parameters: List[str], filename: str, selection_fn_filename: str
) -> PoissonMean:
    """Read the Poisson mean from a JSON file and create a PoissonMean instance.

    Parameters
    ----------
    key : PRNGKeyArray
        The random number generator key.
    parameters : List[str]
        List of parameters to be used in the Poisson mean.
    filename : str
        The path to the JSON file containing the configuration of the Poisson mean.
    selection_fn_filename : str
        The path to the JSON file containing the selection function.

    Returns
    -------
    PoissonMean
        An instance of the PoissonMean class initialized with the configuration from the JSON file.
    """
    selection_fn = vt_json_read_and_process(parameters, selection_fn_filename)
    kwargs = poisson_mean_parser(filename)
    erate_estimator = PoissonMean(
        selection_fn,
        key=key,
        **kwargs,  # type: ignore[arg-type]
    )
    return erate_estimator



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import pandas as pd

import gwkokab
from gwkokab.models import NPowerlawMGaussian
from gwkokab.models.npowerlawmgaussian._ncombination import (
    create_truncated_normal_distributions,
)
from gwkokab.parameters import Parameters
from gwkokab.utils.tools import error_if
from kokab.utils import ppd, ppd_parser
from kokab.utils.common import ppd_ranges, read_json


def make_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = ppd_parser.get_parser(parser)

    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--spin-truncated-normal",
        action="store_true",
        help="Use truncated normal distributions for spin parameters.",
    )

    msm_group = parser.add_argument_group("Multi Source Model Options")
    msm_group.add_argument(
        "--per-component",
        action="store_true",
        help="Compute the PPD for each component of the multi source model.",
    )

    return parser


def compute_per_component_ppd(
    nf_samples,
    ranges,
    constants,
    component_prefix,
    parameters,
    args,
    nf_samples_mapping,
    N_pl,
    N_g,
):
    for n_pl in range(1, N_pl + 1):
        constants_copy = constants.copy()
        constants_copy["N_pl"] = n_pl
        constants_copy["N_g"] = 0
        ppd.compute_and_save_ppd(
            NPowerlawMGaussian,
            nf_samples,
            ranges,
            f"{component_prefix}powerlaw_{n_pl}_" + args.filename,
            parameters,
            constants_copy,
            nf_samples_mapping,
            args.batch_size,
        )

    for n_g in range(1, N_g + 1):
        constants_copy = constants.copy()
        constants_copy["N_pl"] = 0
        constants_copy["N_g"] = n_g
        ppd.compute_and_save_ppd(
            NPowerlawMGaussian,
            nf_samples,
            ranges,
            f"{component_prefix}gaussian_{n_g}_" + args.filename,
            parameters,
            constants_copy,
            nf_samples_mapping,
            args.batch_size,
        )


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()

    if args.spin_truncated_normal:
        gwkokab.models.npowerlawmgaussian._model.build_spin_distributions = (
            create_truncated_normal_distributions
        )

    error_if(
        not str(args.filename).endswith(".hdf5"),
        msg="Output file must be an HDF5 file.",
    )

    constants = read_json(args.constants)
    nf_samples_mapping = read_json(args.nf_samples_mapping)

    N_pl = constants["N_pl"]
    N_g = constants["N_g"]
    has_spin = constants.get("use_spin", False)
    has_tilt = constants.get("use_tilt", False)
    has_eccentricity = constants.get("use_eccentricity", False)
    has_redshift = constants.get("use_redshift", False)
    has_cos_iota = constants.get("use_cos_iota", False)
    has_phi_12 = constants.get("use_phi_12", False)
    has_polarization_angle = constants.get("use_polarization_angle", False)
    has_right_ascension = constants.get("use_right_ascension", False)
    has_sin_declination = constants.get("use_sin_declination", False)
    has_detection_time = constants.get("use_detection_time", False)

    parameters = [
        Parameters.PRIMARY_MASS_SOURCE.value,
        Parameters.SECONDARY_MASS_SOURCE.value,
    ]

    if has_spin:
        parameters.extend(
            [
                Parameters.PRIMARY_SPIN_MAGNITUDE.value,
                Parameters.SECONDARY_SPIN_MAGNITUDE.value,
            ]
        )

    if has_tilt:
        parameters.extend([Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value])

    if has_phi_12:
        parameters.append(Parameters.PHI_12.value)

    if has_eccentricity:
        parameters.append(Parameters.ECCENTRICITY.value)

    if has_redshift:
        parameters.append(Parameters.REDSHIFT.value)

    if has_right_ascension:
        parameters.append(Parameters.RIGHT_ASCENSION.value)

    if has_sin_declination:
        parameters.append(Parameters.SIN_DECLINATION.value)

    if has_detection_time:
        parameters.append(Parameters.DETECTION_TIME.value)

    if has_cos_iota:
        parameters.append(Parameters.COS_IOTA.value)

    if has_polarization_angle:
        parameters.append(Parameters.POLARIZATION_ANGLE.value)

    ranges = ppd_ranges(parameters, args.range)

    nf_samples = pd.read_csv(
        args.sample_filename, delimiter=" ", comment="#", header=None
    ).to_numpy()

    ppd.compute_and_save_ppd(
        NPowerlawMGaussian,
        nf_samples,
        ranges,
        "rate_scaled_" + args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )

    if args.per_component:
        compute_per_component_ppd(
            nf_samples,
            ranges,
            constants,
            "rate_scaled_",
            parameters,
            args,
            nf_samples_mapping,
            N_pl,
            N_g,
        )

    nf_samples, constants = ppd.wipe_log_rate(nf_samples, nf_samples_mapping, constants)

    ppd.compute_and_save_ppd(
        NPowerlawMGaussian,
        nf_samples,
        ranges,
        args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )

    if args.per_component:
        compute_per_component_ppd(
            nf_samples,
            ranges,
            constants,
            "",
            parameters,
            args,
            nf_samples_mapping,
            N_pl,
            N_g,
        )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentParser
from collections.abc import Callable
from typing import List, Tuple, Union

import h5py
import jax
from jax import numpy as jnp
from jaxtyping import Array
from loguru import logger
from numpyro.distributions import Distribution
from numpyro.distributions.distribution import enable_validation

from gwkokab.inference import analytical_likelihood
from gwkokab.utils.tools import error_if
from kokab.utils.poisson_mean_parser import read_pmean

from .flowMC_based import FlowMCBased
from .guru import guru_arg_parser as guru_parser


def _read_mean_covariances(filename: str) -> Tuple[List[Array], List[Array]]:
    """Reads means and covariances from a file.

    Args:
        filename (str): The path to the file containing means and covariances.

    Returns:
        Tuple[List[Array], List[Array]]: A tuple containing two lists:
            - list_of_means: List of mean arrays.
            - list_of_covariances: List of covariance arrays.
    """
    assert filename.endswith(".hdf5") or filename.endswith(".h5"), (
        "The filename must end with '.hdf5' or '.h5'."
    )

    logger.info("Reading means and covariances from {filename}", filename=filename)

    list_of_means = []
    list_of_covariances = []

    with h5py.File(filename, "r") as f:
        for key in f.keys():
            if not key.startswith("event_"):
                continue

            group = f[key]
            error_if(
                "mean" not in group,
                msg=f"Key 'mean' not found in group {key} of file {filename}.",
            )
            error_if(
                "cov" not in group,
                msg=f"Key 'cov' not found in group {key} of file {filename}.",
            )

            mean = group["mean"][:]
            cov = group["cov"][:]

            n_dim = mean.shape[0]
            error_if(
                cov.shape != (n_dim, n_dim),
                msg=f"Covariance shape {cov.shape} does not match mean shape "
                f"{mean.shape} in group {key} of file {filename}.",
            )

            list_of_means.append(mean)
            list_of_covariances.append(cov)

    return list_of_means, list_of_covariances


class Monk(FlowMCBased):
    def __init__(
        self,
        model: Union[Distribution, Callable[..., Distribution]],
        data_filename: str,
        seed: int,
        prior_filename: str,
        selection_fn_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_samples: int,
        max_iter_mean: int,
        max_iter_cov: int,
        n_vi_steps: int,
        learning_rate: float,
        batch_size: int,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
        analysis_name: str = "",
    ) -> None:
        """

        Parameters
        ----------
        model : Union[Distribution, Callable[..., Distribution]]
            model to be used in the Monk class. It can be a Distribution or a callable
            that returns a Distribution.
        data_filename : str
            path to the HDF5 file containing the data.
        seed : int
            seed for the random number generator.
        prior_filename : str
            path to the JSON file containing the prior distributions.
        selection_fn_filename : str
            path to the JSON file containing the selection function.
        poisson_mean_filename : str
            path to the JSON file containing the Poisson mean configuration.
        flowMC_settings_filename : str
            path to the JSON file containing the flowMC settings.
        debug_nans : bool, optional
            If True, checks for NaNs in each computation. See details in the
            [documentation](https://jax.readthedocs.io/en/latest/_autosummary/jax.debug_nans.html#jax.debug_nans),
            by default False
        profile_memory : bool, optional
            If True, enables memory profiling, by default False
        check_leaks : bool, optional
            If True, checks for JAX Tracer leaks. See details in the
            [documentation](https://jax.readthedocs.io/en/latest/_autosummary/jax.checking_leaks.html#jax.checking_leaks),
            by default False
        analysis_name : str, optional
            Name of the analysis, by default ""
        """
        assert all(letter.isalpha() or letter == "_" for letter in analysis_name), (
            "Analysis name must be alphabetic characters only."
        )
        self.data_filename = data_filename
        self.n_samples = n_samples
        self.max_iter_mean = max_iter_mean
        self.max_iter_cov = max_iter_cov
        self.n_vi_steps = n_vi_steps
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.set_rng_key(seed=seed)

        super().__init__(
            analysis_name=analysis_name or model.__name__,
            check_leaks=check_leaks,
            debug_nans=debug_nans,
            model=model,
            poisson_mean_filename=poisson_mean_filename,
            prior_filename=prior_filename,
            profile_memory=profile_memory,
            sampler_settings_filename=sampler_settings_filename,
            selection_fn_filename=selection_fn_filename,
        )

    def run(self) -> None:
        """Runs the Monk analysis."""
        _, dist_fn, priors, variables, variables_index = self.bake_model()

        list_of_means, list_of_covariances = _read_mean_covariances(self.data_filename)

        n_events = len(list_of_means)
        mean_stack: Array = jax.block_until_ready(
            jax.device_put(jnp.stack(list_of_means, axis=0), may_alias=True)
        )
        cov_stack: Array = jax.block_until_ready(
            jax.device_put(jnp.stack(list_of_covariances, axis=0), may_alias=True)
        )

        logger.debug("mean_stack.shape: {shape}", shape=mean_stack.shape)
        logger.debug("cov_stack.shape: {shape}", shape=cov_stack.shape)

        ERate_fn = read_pmean(
            self.rng_key,
            self.parameters,
            self.poisson_mean_filename,
            self.selection_fn_filename,
        )

        logpdf = analytical_likelihood(
            dist_fn,
            priors,
            variables_index,
            ERate_fn,
            self.rng_key,
            n_events=n_events,
            n_samples=self.n_samples,
            max_iter_mean=self.max_iter_mean,
            max_iter_cov=self.max_iter_cov,
            n_vi_steps=self.n_vi_steps,
            learning_rate=self.learning_rate,
            batch_size=self.batch_size,
        )

        self.driver(
            logpdf=logpdf,
            priors=priors,
            data={"mean_stack": mean_stack, "cov_stack": cov_stack},
            labels=sorted(variables.keys()),
        )


def monk_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Monk script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    # Global enable validation for all distributions
    enable_validation()

    parser = guru_parser(parser)

    monk_group = parser.add_argument_group("Monk Options")
    monk_group.add_argument(
        "--data-filename",
        help="Path to the HDF5 file containing the data.",
        type=str,
        required=True,
    )

    likelihood_group = parser.add_argument_group("Likelihood Options")
    likelihood_group.add_argument(
        "--n-samples",
        help="Number of samples to draw from the multivariate normal distribution for each "
        "event to compute the likelihood, by default 10_000",
        default=10_000,
        type=int,
    )
    likelihood_group.add_argument(
        "--max-iter-mean",
        help="Maximum number of iterations for the fitting process of the mean, by default 10",
        default=10,
        type=int,
    )
    likelihood_group.add_argument(
        "--max-iter-cov",
        help="Maximum number of iterations for the fitting process of the covariance, by default 3",
        default=3,
        type=int,
    )
    likelihood_group.add_argument(
        "--n-vi-steps",
        help="Number of steps for the variational inference",
        default=5,
        type=int,
    )
    likelihood_group.add_argument(
        "--learning-rate",
        help="Learning rate for the variational inference, by default 0.01",
        default=0.01,
        type=float,
    )
    likelihood_group.add_argument(
        "--batch-size",
        help="Batch size for the `jax.lax.map` used in the likelihood computation, by default 1000",
        default=1_000,
        type=int,
    )

    return parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Dict, List, Literal, Optional

from jaxtyping import Array
from numpyro.distributions import (
    Beta,
    Distribution,
    MixtureGeneral,
    Normal,
    TruncatedNormal,
    TwoSidedTruncatedDistribution,
    Uniform,
)

from ...utils.tools import fetch_first_matching_value
from .._models import PowerlawPrimaryMassRatio
from ..redshift import PowerlawRedshift
from ..spin import BetaFromMeanVar, IndependentSpinOrientationGaussianIsotropic
from ..transformations import PrimaryMassAndMassRatioToComponentMassesTransform
from ..utils import ExtendedSupportTransformedDistribution


__all__ = [
    "combine_distributions",
    "create_beta_distributions",
    "create_powerlaw_redshift",
    "create_powerlaws",
    "create_truncated_normal_distributions_for_cos_tilt",
    "create_truncated_normal_distributions",
]


def combine_distributions(
    base_dists: List[List[Distribution]], add_dists: List[Distribution]
):
    """Helper function to combine base distributions with additional distributions like
    spin, tilt, or eccentricity.
    """
    return [dists + [add_dist] for dists, add_dist in zip(base_dists, add_dists)]


def create_beta_distributions(
    N: int,
    parameter_name: Literal["chi1", "chi2"],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Beta]:
    """Create a list of Beta distributions.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[&quot;chi1&quot;, &quot;chi2&quot;]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[Beta]
        list of Beta distributions

    Raises
    ------
    ValueError
        if mean or variance is missing
    """

    beta_collection = []
    mean_name = f"{parameter_name}_mean_{component_type}"
    variance_name = f"{parameter_name}_variance_{component_type}"
    for i in range(N):
        mean = fetch_first_matching_value(params, f"{mean_name}_{i}", mean_name)
        if mean is None:
            raise ValueError(f"Missing parameter {mean_name}_{i}")
        variance = fetch_first_matching_value(
            params, f"{variance_name}_{i}", variance_name
        )
        if variance is None:
            raise ValueError(f"Missing parameter {variance_name}_{i}")

        beta_collection.append(
            BetaFromMeanVar(
                mean=mean,
                variance=variance,
                validate_args=validate_args,
            )
        )
    return beta_collection


def create_truncated_normal_distributions(
    N: int,
    parameter_name: Literal[
        "m1",
        "m2",
        "chi1",
        "chi2",
        "cos_tilt1",
        "cos_tilt2",
        "ecc",
        "cos_iota",
        "phi_12",
        "polarization_angle",
        "right_ascension",
        "sin_declination",
    ],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Distribution]:
    """Create a list of TruncatedNormal distributions.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[ &quot;m1&quot;, &quot;m2&quot;, &quot;chi1&quot;, &quot;chi2&quot;, &quot;cos_tilt1&quot;, &quot;cos_tilt2&quot;, &quot;ecc&quot;, &quot;cos_iota&quot;, &quot;phi_12&quot;, &quot;polarization_angle&quot;, &quot;right_ascension&quot;, &quot;sin_declination&quot;, ]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[Distribution]
        list of TruncatedNormal distributions

    Raises
    ------
    ValueError
        if loc, scale, low, or high is missing
    """
    truncated_normal_collection = []
    loc_name = f"{parameter_name}_loc_{component_type}"
    scale_name = f"{parameter_name}_scale_{component_type}"
    low_name = f"{parameter_name}_low_{component_type}"
    high_name = f"{parameter_name}_high_{component_type}"
    for i in range(N):
        loc = fetch_first_matching_value(params, f"{loc_name}_{i}", loc_name)
        if loc is None:
            raise ValueError(f"Missing parameter {loc_name}_{i}")

        scale = fetch_first_matching_value(params, f"{scale_name}_{i}", scale_name)
        if scale is None:
            raise ValueError(f"Missing parameter {scale_name}_{i}")

        low = fetch_first_matching_value(params, f"{low_name}_{i}", low_name)
        if low is None:
            raise ValueError(f"Missing parameter {low_name}_{i}")

        high = fetch_first_matching_value(params, f"{high_name}_{i}", high_name)
        if high is None:
            raise ValueError(f"Missing parameter {high_name}_{i}")

        truncated_normal_collection.append(
            TruncatedNormal(
                loc=loc, scale=scale, low=low, high=high, validate_args=validate_args
            )
        )

    return truncated_normal_collection


def create_truncated_normal_distributions_for_cos_tilt(
    N: int,
    parameter_name: Literal["cos_tilt1", "cos_tilt2"],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[TwoSidedTruncatedDistribution]:
    """Create a list of TwoSidedTruncatedDistribution distributions for tilt.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[&quot;cos_tilt1&quot;, &quot;cos_tilt2&quot;]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[TwoSidedTruncatedDistribution]
        list of TwoSidedTruncatedDistribution distributions

    Raises
    ------
    ValueError
        if scale is missing
    """
    truncated_normal_for_tilt_collection = []
    scale_name = f"{parameter_name}_scale_{component_type}"
    for i in range(N):
        scale = fetch_first_matching_value(params, f"{scale_name}_{i}", scale_name)
        if scale is None:
            raise ValueError(f"Missing parameter {scale_name}_{i}")

        truncated_normal_for_tilt_collection.append(
            TwoSidedTruncatedDistribution(
                Normal(loc=1.0, scale=scale, validate_args=validate_args),
                low=-1,
                high=1,
                validate_args=validate_args,
            )
        )

    return truncated_normal_for_tilt_collection


def create_independent_spin_orientation_gaussian_isotropic(
    N: int,
    parameter_name: Literal["cos_tilt1", "cos_tilt2"],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[MixtureGeneral]:
    """Create a list of :func:`IndependentSpinOrientationGaussianIsotropic`
    distributions for tilt.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[&quot;cos_tilt1&quot;, &quot;cos_tilt2&quot;]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[MixtureGeneral]
        list of :func:`IndependentSpinOrientationGaussianIsotropic` distributions

    Raises
    ------
    ValueError
        if scale is missing
    """
    dist_collection = []
    zeta_name = f"cos_tilt_zeta_{component_type}"
    scale1_name = f"cos_tilt1_scale_{component_type}"
    scale2_name = f"cos_tilt2_scale_{component_type}"

    for i in range(N):
        zeta = fetch_first_matching_value(params, f"{zeta_name}_{i}", zeta_name)
        if zeta is None:
            raise ValueError(f"Missing parameter {zeta_name}_{i}")

        scale1 = fetch_first_matching_value(params, f"{scale1_name}_{i}", scale1_name)
        if scale1 is None:
            raise ValueError(f"Missing parameter {scale1_name}_{i}")

        scale2 = fetch_first_matching_value(params, f"{scale2_name}_{i}", scale2_name)
        if scale2 is None:
            raise ValueError(f"Missing parameter {scale2_name}_{i}")

        dist_collection.append(
            IndependentSpinOrientationGaussianIsotropic(
                zeta=zeta,
                scale1=scale1,
                scale2=scale2,
                validate_args=validate_args,
            )
        )

    return dist_collection


def create_powerlaws(
    N: int,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[ExtendedSupportTransformedDistribution]:
    """Create a list of ExtendedSupportTransformedDistribution for powerlaws.

    Parameters
    ----------
    N : int
        Number of components
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[ExtendedSupportTransformedDistribution]
        list of ExtendedSupportTransformedDistribution for powerlaws

    Raises
    ------
    ValueError
        if alpha, beta, mmin, or mmax is missing
    """
    powerlaws_collection = []
    alpha_name = "alpha_pl"
    beta_name = "beta_pl"
    mmin_name = "mmin_pl"
    mmax_name = "mmax_pl"
    for i in range(N):
        alpha = fetch_first_matching_value(params, f"{alpha_name}_{i}", alpha_name)
        if alpha is None:
            raise ValueError(f"Missing parameter {alpha_name}_{i}")

        beta = fetch_first_matching_value(params, f"{beta_name}_{i}", beta_name)
        if beta is None:
            raise ValueError(f"Missing parameter {beta_name}_{i}")

        mmin = fetch_first_matching_value(params, f"{mmin_name}_{i}", mmin_name)
        if mmin is None:
            raise ValueError(f"Missing parameter {mmin_name}_{i}")

        mmax = fetch_first_matching_value(params, f"{mmax_name}_{i}", mmax_name)
        if mmax is None:
            raise ValueError(f"Missing parameter {mmax_name}_{i}")

        powerlaw = PowerlawPrimaryMassRatio(
            alpha=alpha, beta=beta, mmin=mmin, mmax=mmax, validate_args=validate_args
        )
        transformed_powerlaw = ExtendedSupportTransformedDistribution(
            base_distribution=powerlaw,
            transforms=PrimaryMassAndMassRatioToComponentMassesTransform(),
            validate_args=validate_args,
        )
        powerlaws_collection.append(transformed_powerlaw)
    return powerlaws_collection


def create_powerlaw_redshift(
    N: int,
    parameter_name: Literal["redshift"],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Distribution]:
    """Create a list of PowerlawRedshift distributions.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[&quot;redshift&quot;]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None, by default None

    Returns
    -------
    List[Distribution]
        list of PowerlawRedshift distributions

    Raises
    ------
    ValueError
        if :code:`kappa` or :code:`z_max` parameters are missing
    """
    powerlaw_redshift_collection = []
    kappa_name = f"{parameter_name}_kappa_{component_type}"
    z_max_name = f"{parameter_name}_z_max_{component_type}"

    for i in range(N):
        kappa = fetch_first_matching_value(params, f"{kappa_name}_{i}", kappa_name)
        if kappa is None:
            raise ValueError(f"Missing parameter {kappa_name}_{i}")

        z_max = fetch_first_matching_value(params, f"{z_max_name}_{i}", z_max_name)
        if z_max is None:
            raise ValueError(f"Missing parameter {z_max_name}_{i}")

        powerlaw_redshift_collection.append(
            PowerlawRedshift(kappa=kappa, z_max=z_max, validate_args=validate_args)
        )

    return powerlaw_redshift_collection


def create_uniform_distributions(
    N: int,
    parameter_name: Literal[
        "chi1",
        "chi2",
        "cos_iota",
        "cos_tilt1",
        "cos_tilt2",
        "dec",
        "detection_time",
        "ecc",
        "m1",
        "m2",
        "mean_anomaly",
        "phi_1",
        "phi_12",
        "phi_2",
        "phi_orb",
        "psi",
        "ra",
        "redshift",
    ],
    component_type: Literal["pl", "g"],
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Distribution]:
    """Create a list of Uniform distributions.

    Parameters
    ----------
    N : int
        Number of components
    parameter_name : Literal[ &quot;chi1&quot;, &quot;chi2&quot;, &quot;cos_iota&quot;, &quot;cos_tilt1&quot;, &quot;cos_tilt2&quot;, &quot;dec&quot;, &quot;detection_time&quot;, &quot;ecc&quot;, &quot;m1&quot;, &quot;m2&quot;, &quot;mean_anomaly&quot;, &quot;phi_1&quot;, &quot;phi_12&quot;, &quot;phi_2&quot;, &quot;phi_orb&quot;, &quot;psi&quot;, &quot;ra&quot;, &quot;redshift&quot;, ]
        name of the parameter to create distributions for
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either &quot;pl&quot; or &quot;g&quot;
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[Distribution]
        list of Uniform distributions
    """
    uniform_collection = []
    low_name = f"{parameter_name}_low_{component_type}"
    high_name = f"{parameter_name}_high_{component_type}"
    for i in range(N):
        low = fetch_first_matching_value(params, f"{low_name}_{i}", low_name)
        if low is None:
            raise ValueError(f"Missing parameter {low_name}_{i}")

        high = fetch_first_matching_value(params, f"{high_name}_{i}", high_name)
        if high is None:
            raise ValueError(f"Missing parameter {high_name}_{i}")

        uniform_collection.append(
            Uniform(low=low, high=high, validate_args=validate_args)
        )

    return uniform_collection



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import sys

from loguru import logger


def set_log_level(log_level: str) -> None:
    """Set the log level for the logger. The preset log level when initialising GWKokab
    is the value of the GWKOKAB_LOG_LEVEL environment variable, or 'WARNING' if the
    environment variable is unset.

    Parameters
    ----------
    log_level : str
        The log level to set. Options are 'TRACE','DEBUG', 'INFO', 'SUCCESS', 'WARNING',
        'ERROR', 'CRITICAL'
    """
    logger.remove()
    logger.add(
        sys.stderr,
        level=log_level,
        format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        "<level>{message}</level>",
    )
    logger.debug(f"Setting LogLevel to {log_level}")


def log_gwkokab_info() -> None:
    """Prints the GWKokab version and the Python version."""

    import gwkokab as gwk

    logger.info("=" * 60)
    logger.info("GWKokab INFO")
    logger.info("=" * 60)

    logger.info("GWKokab version: {gwk_version}", gwk_version=gwk.__version__)
    logger.info("Python version: {python_version}", python_version=sys.version)
    logger.info("Python platform: {python_platform}", python_platform=sys.platform)
    logger.info("Python build: {python_build}", python_build=sys.version_info)

    del gwk


def log_device_info() -> None:
    """Prints the device information."""
    import jax as _jax
    import jaxlib as _jaxlib
    from jax.extend.backend import get_backend

    logger.info("=" * 60)
    logger.info("JAX CUDA ENVIRONMENT INFO")
    logger.info("=" * 60)

    logger.info("Devices count: {n_devices}", n_devices=_jax.device_count())
    logger.info("Devices: {devices}", devices=_jax.devices())

    logger.info("jax version: {jax_version}", jax_version=_jax.__version__)
    logger.info("jaxlib version: {jaxlib_version}", jaxlib_version=_jaxlib.__version__)

    try:
        backend = get_backend()
        logger.info("JAX platform: {}", backend.platform)
        logger.info("JAX backend: {}", type(backend).__name__)

        platform_version = getattr(backend, "platform_version", None)
        if platform_version:
            logger.info("JAX platform version: {}", platform_version)
        else:
            logger.warning("No platform version info found on backend.")
    except Exception as e:
        logger.warning("Could not retrieve CUDA info from XLA backend: {}", e)

    for device in _jax.devices():
        logger.info(
            "Device: {}, process_index: {}, id: {}, platform: {}",
            device.device_kind,
            device.process_index,
            device.id,
            device.platform,
        )

    del _jax
    del _jaxlib
    del get_backend


def log_info(start: bool = False) -> None:
    """Log the information about the package and the device information.

    Parameters
    ----------
    start : bool, optional
        If True, log the start message, by default False
    """
    log_gwkokab_info()
    log_device_info()
    if start:
        logger.info("=" * 60)
        logger.info("GWKokab STARTING")
        logger.info("=" * 60)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""Provides essential classes and functions for the inference module."""

from .analyticallikelihood import analytical_likelihood as analytical_likelihood
from .poissonlikelihood import poisson_likelihood as poisson_likelihood
from .poissonlikelihood_numpyro import (
    numpyro_poisson_likelihood as numpyro_poisson_likelihood,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
from jax import numpy as jnp

from gwkokab.utils.transformations import (
    cart_to_polar,
    cart_to_spherical,
    chi_costilt_to_chiz,
    chieff,
    chirp_mass,
    delta_m,
    delta_m_to_symmetric_mass_ratio,
    m1_m2_chi1_chi2_costilt1_costilt2_to_chieff,
    m1_m2_chi1_chi2_costilt1_costilt2_to_chiminus,
    m1_m2_chi1z_chi2z_to_chiminus,
    m1_m2_chieff_chiminus_to_chi1z_chi2z,
    m1_q_to_m2,
    m1_times_m2,
    m2_q_to_m1,
    m_det_z_to_m_source,
    m_source_z_to_m_det,
    mass_ratio,
    Mc_eta_to_m1_m2,
    polar_to_cart,
    reduced_mass,
    spherical_to_cart,
    symmetric_mass_ratio,
    symmetric_mass_ratio_to_delta_m,
    total_mass,
)


_primary_masses = np.unique(np.random.uniform(0.5, 200, 13))
_secondary_masses = np.array([np.random.uniform(0.5, m) for m in _primary_masses])


_m1m2 = [(m1, m2) for m1, m2 in zip(_primary_masses, _secondary_masses)]


_redshifts = np.random.uniform(0, 1, 3)

_rho = np.random.uniform(0, 10, 3)
_thetas = np.random.uniform(0, 2 * np.pi, 3)
_phis = np.random.uniform(0, np.pi, 3)

_x = np.random.uniform(-10, 10, 7)
_y = np.random.uniform(-10, 10, 7)
_z = np.random.uniform(-10, 10, 7)

_chi = np.random.uniform(-1, 1, 3)
_chi1 = np.random.uniform(-1, 1, 3)
_chi2 = np.random.uniform(-1, 1, 3)

_costilt = np.random.uniform(-1, 1, 3)
_costilt1 = np.random.uniform(-1, 1, 3)
_costilt2 = np.random.uniform(-1, 1, 3)


@pytest.mark.parametrize("m1, m2", _m1m2)
def test_different_mass_representations(m1, m2):
    m1m2 = m1 * m2
    assert jnp.allclose(m1m2, m1_times_m2(m1=m1, m2=m2))
    M = m1 + m2
    assert jnp.allclose(M, total_mass(m1=m1, m2=m2))
    q = m2 / m1
    assert jnp.allclose(q, mass_ratio(m1=m1, m2=m2))
    assert jnp.allclose(m1, m2_q_to_m1(m2=m2, q=q))
    assert jnp.allclose(m2, m1_q_to_m2(m1=m1, q=q))
    Mc = m1m2**0.6 * M**-0.2
    assert jnp.allclose(Mc, chirp_mass(m1=m1, m2=m2))
    eta = m1m2 * M**-2.0
    assert jnp.allclose(eta, symmetric_mass_ratio(m1=m1, m2=m2))
    Mr = m1m2 / M
    assert jnp.allclose(Mr, reduced_mass(m1=m1, m2=m2))
    delta = (m1 - m2) / M
    assert jnp.allclose(delta, delta_m(m1=m1, m2=m2))
    eta_ = (1 - jnp.square(delta)) / 4
    assert jnp.allclose(eta_, delta_m_to_symmetric_mass_ratio(delta_m=delta))
    delta_ = jnp.sqrt(1 - 4 * eta)
    assert jnp.allclose(delta_, symmetric_mass_ratio_to_delta_m(eta=eta))
    m1_, m2_ = Mc_eta_to_m1_m2(Mc=Mc, eta=eta)
    assert jnp.allclose(m1, m1_)
    assert jnp.allclose(m2, m2_)


@pytest.mark.parametrize("m", _primary_masses)
@pytest.mark.parametrize("z", _redshifts)
def test_mass_and_reshift(m, z):
    m_source = m / (1.0 + z)
    assert jnp.allclose(m_source, m_det_z_to_m_source(m_det=m, z=z))
    m_det = m * (1.0 + z)
    assert jnp.allclose(m_det, m_source_z_to_m_det(m_source=m, z=z))


@pytest.mark.parametrize("chi", _chi)
@pytest.mark.parametrize("costilt", _costilt)
def test_chi_costilt_to_chiz(chi, costilt):
    chi_z = chi * costilt
    assert jnp.allclose(chi_z, chi_costilt_to_chiz(chi=chi, costilt=costilt))


@pytest.mark.parametrize("m1, m2", _m1m2)
@pytest.mark.parametrize("chi1", _chi1)
@pytest.mark.parametrize("chi2", _chi2)
def test_m1_m2_chi1z_chi2z(m1, m2, chi1, chi2):
    chieff_ = chieff(m1=m1, m2=m2, chi1z=chi1, chi2z=chi2)
    chieff__ = (m1 * chi1 + m2 * chi2) / (m1 + m2)
    assert jnp.allclose(chieff__, chieff_)
    chiminus_ = m1_m2_chi1z_chi2z_to_chiminus(m1=m1, m2=m2, chi1z=chi1, chi2z=chi2)
    chiminus = (m1 * chi1 - m2 * chi2) / (m1 + m2)
    assert jnp.allclose(chiminus, chiminus_)


@pytest.mark.parametrize("m1, m2", _m1m2)
@pytest.mark.parametrize("chieff", _chi)
@pytest.mark.parametrize("chiminus", _chi)
def test_m1_m2_chieff_chiminus(m1, m2, chieff, chiminus):
    chi1z_, chi2z_ = m1_m2_chieff_chiminus_to_chi1z_chi2z(
        m1=m1, m2=m2, chieff=chieff, chiminus=chiminus
    )
    chi1z = (m1 + m2) * (chieff + chiminus) / (2 * m1)
    chi2z = (m1 + m2) * (chieff - chiminus) / (2 * m2)
    assert jnp.allclose(chi1z, chi1z_)
    assert jnp.allclose(chi2z, chi2z_)


@pytest.mark.parametrize("m1, m2", _m1m2)
@pytest.mark.parametrize("chi1", _chi1)
@pytest.mark.parametrize("chi2", _chi2)
@pytest.mark.parametrize("costilt1", _costilt1)
@pytest.mark.parametrize("costilt2", _costilt2)
def test_m1_m2_chi1_chi2_costilt1_costilt2(m1, m2, chi1, chi2, costilt1, costilt2):
    chieff_ = m1_m2_chi1_chi2_costilt1_costilt2_to_chieff(
        m1=m1, m2=m2, chi1=chi1, chi2=chi2, costilt1=costilt1, costilt2=costilt2
    )
    chieff = (m1 * chi1 * costilt1 + m2 * chi2 * costilt2) / (m1 + m2)
    assert jnp.allclose(chieff, chieff_)
    chiminus_ = m1_m2_chi1_chi2_costilt1_costilt2_to_chiminus(
        m1=m1, m2=m2, chi1=chi1, chi2=chi2, costilt1=costilt1, costilt2=costilt2
    )
    chiminus = (m1 * chi1 * costilt1 - m2 * chi2 * costilt2) / (m1 + m2)
    assert jnp.allclose(chiminus, chiminus_)


@pytest.mark.parametrize("x", _x)
@pytest.mark.parametrize("y", _y)
def test_cart_to_polar(x, y):
    r, theta = cart_to_polar(x=x, y=y)
    r_ = jnp.sqrt(jnp.square(x) + jnp.square(y))
    theta_ = jnp.arctan2(y, x)
    assert jnp.allclose(r, r_)
    assert jnp.allclose(theta, theta_)


@pytest.mark.parametrize("r", _rho)
@pytest.mark.parametrize("theta", _thetas)
def test_polar_to_cart(r, theta):
    x, y = polar_to_cart(r=r, theta=theta)
    x_ = r * jnp.cos(theta)
    y_ = r * jnp.sin(theta)
    assert jnp.allclose(x, x_)
    assert jnp.allclose(y, y_)


@pytest.mark.parametrize("r", _rho)
@pytest.mark.parametrize("theta", _thetas)
@pytest.mark.parametrize("phi", _phis)
def test_spherical_to_cart(r, theta, phi):
    x, y, z = spherical_to_cart(r=r, theta=theta, phi=phi)
    x_ = r * jnp.sin(theta) * jnp.cos(phi)
    y_ = r * jnp.sin(theta) * jnp.sin(phi)
    z_ = r * jnp.cos(theta)
    assert jnp.allclose(x, x_)
    assert jnp.allclose(y, y_)
    assert jnp.allclose(z, z_)


@pytest.mark.parametrize("x", _x)
@pytest.mark.parametrize("y", _y)
@pytest.mark.parametrize("z", _z)
def test_cart_to_spherical(x, y, z):
    r, theta, phi = cart_to_spherical(x=x, y=y, z=z)
    r_ = jnp.sqrt(jnp.square(x) + jnp.square(y) + jnp.square(z))
    theta_ = jnp.arccos(z / r_)
    phi_ = jnp.arctan2(y, x)
    assert jnp.allclose(r, r_)
    assert jnp.allclose(theta, theta_)
    assert jnp.allclose(phi, phi_)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable
from typing import Any, Dict, List, Optional, Tuple

import equinox as eqx
import jax
from jax import Array, numpy as jnp
from jaxtyping import ArrayLike
from numpyro._typing import DistributionLike

from ..models.utils import JointDistribution
from ..poisson_mean import PoissonMean


__all__ = ["poisson_likelihood"]


def poisson_likelihood(
    dist_fn: Callable[..., DistributionLike],
    priors: JointDistribution,
    variables: Dict[str, DistributionLike],
    variables_index: Dict[str, int],
    log_constants: ArrayLike,
    ERate_obj: PoissonMean,
    where_fns: Optional[List[Callable[..., Array]]],
    constants: Dict[str, Array],
) -> Callable[[Array, Dict[str, Any]], Array]:
    r"""This class is used to provide a likelihood function for the inhomogeneous Poisson
    process. The likelihood is given by,

    .. math::
        \log\mathcal{L}(\Lambda) \propto -\mu(\Lambda)
        +\log\sum_{n=1}^N \int \ell_n(\lambda) \rho(\lambda\mid\Lambda)
        \mathrm{d}\lambda


    where, :math:`\displaystyle\rho(\lambda\mid\Lambda) =
    \frac{\mathrm{d}N}{\mathrm{d}V\mathrm{d}t \mathrm{d}\lambda}` is the merger
    rate density for a population parameterized by :math:`\Lambda`, :math:`\mu(\Lambda)` is
    the expected number of detected mergers for that population, and
    :math:`\ell_n(\lambda)` is the likelihood for the :math:`n`-th observed event's
    parameters. Using Bayes' theorem, we can obtain the posterior
    :math:`p(\Lambda\mid\text{data})` by multiplying the likelihood by a prior
    :math:`\pi(\Lambda)`.

    .. math::
        p(\Lambda\mid\text{data}) \propto \pi(\Lambda) \mathcal{L}(\Lambda)

    The integral inside the main likelihood expression is then evaluated via
    Monte Carlo as

    .. math::
        \int \ell_n(\lambda) \rho(\lambda\mid\Lambda) \mathrm{d}\lambda \propto
        \int \frac{p(\lambda | \mathrm{data}_n)}{\pi_n(\lambda)}
        \rho(\lambda\mid\Lambda) \mathrm{d}\lambda \approx
        \frac{1}{N_{\mathrm{samples}}}
        \sum_{i=1}^{N_{\mathrm{samples}}}
        \frac{\rho(\lambda_{n,i}\mid\Lambda)}{\pi_{n,i}}
    """
    del variables

    def likelihood_fn(x: Array, data: Dict[str, Tuple[Array]]) -> Array:
        data_group: Tuple[Array] = data["data_group"]
        log_ref_priors_group: Tuple[Array] = data["log_ref_priors_group"]
        masks_group: Tuple[Array] = data["masks_group"]
        mapped_params = {
            name: jax.lax.dynamic_index_in_dim(x, i, keepdims=False)
            for name, i in variables_index.items()
        }

        model_instance = dist_fn(**mapped_params)

        # μ = E_{Ω|Λ}[VT(ω)]
        expected_rates = ERate_obj(model_instance)

        log_prob_fn = eqx.filter_jit(eqx.filter_vmap(model_instance.log_prob))

        def single_event_fn(
            carry: Array, input: Tuple[Array, Array, Array]
        ) -> Tuple[Array, None]:
            data, log_ref_prior, mask = input

            safe_data = jnp.where(
                jnp.expand_dims(mask, axis=-1),
                data,
                model_instance.support.feasible_like(data),
            )
            safe_log_ref_prior = jnp.where(mask, log_ref_prior, 0.0)

            # log p(ω|data_n)
            model_log_prob = log_prob_fn(safe_data)
            safe_model_log_prob = jnp.where(mask, model_log_prob, -jnp.inf)

            # log p(ω|data_n) - log π_n
            log_prob: Array = safe_model_log_prob - safe_log_ref_prior
            log_prob = jnp.where(mask & (~jnp.isnan(log_prob)), log_prob, -jnp.inf)

            # log Σ exp (log p(ω|data_n) - log π_n)
            log_prob_sum = jax.nn.logsumexp(
                log_prob,
                axis=-1,
                where=(~jnp.isneginf(log_prob)) & mask,
            )
            return carry + log_prob_sum, None

        total_log_likelihood = log_constants  # - Σ log(M_i)
        # Σ log Σ exp (log p(ω|data_n) - log π_n)
        for batched_data, batched_log_ref_priors, batched_mask in zip(
            data_group, log_ref_priors_group, masks_group
        ):
            total_log_likelihood, _ = jax.lax.scan(
                single_event_fn,  # type: ignore
                total_log_likelihood,
                (batched_data, batched_log_ref_priors, batched_mask),
            )

        log_prior = priors.log_prob(x)
        # log L(ω) = -μ + Σ log Σ exp (log p(ω|data_n) - log π_n) - Σ log(M_i)
        log_likelihood = total_log_likelihood - expected_rates
        # log p(ω|data) = log π(ω) + log L(ω)
        log_posterior = log_prior + log_likelihood

        log_posterior = jnp.nan_to_num(
            log_posterior,
            nan=-jnp.inf,
            posinf=-jnp.inf,
            neginf=-jnp.inf,
        )

        return log_posterior

    if where_fns is None:
        return eqx.filter_jit(likelihood_fn)

    def likelihood_fn_with_checks(x: Array, data: Dict[str, Tuple[Array]]) -> Array:
        mapped_params = {
            name: jax.lax.dynamic_index_in_dim(x, i, keepdims=False)
            for name, i in variables_index.items()
        }
        predicate = priors.support.check(x)
        for where_fn in where_fns:  # type: ignore
            predicate = jnp.logical_and(
                predicate, where_fn(**constants, **mapped_params)
            )
        predicate = jnp.logical_and(jnp.all(jnp.isfinite(x)), predicate)
        return jnp.where(predicate, likelihood_fn(x, data), -jnp.inf)

    return eqx.filter_jit(likelihood_fn_with_checks)



# Copyright (c) 2025 Tom Callister
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json
from abc import abstractmethod
from collections.abc import Callable
from typing import Optional, Tuple

import equinox as eqx
import h5py
import jax
from jax import numpy as jnp
from jaxtyping import Array, PRNGKeyArray

from ...utils.tools import warn_if
from .._abc import VolumeTimeSensitivityInterface


class Emulator(VolumeTimeSensitivityInterface):
    """Base class implementing a generic detection probability emulator.

    Intended to be subclassed when constructing emulators for particular
    networks/observing runs.
    """

    nn_vt: eqx.nn.MLP = eqx.field()
    scaler: dict[str, Array] = eqx.field()
    scale: float = eqx.field()

    def __init__(
        self,
        trained_weights: str,
        scaler: str,
        input_size: int,
        hidden_layer_width: int,
        hidden_layer_depth: int,
        activation: Callable,
        final_activation: Callable,
        batch_size: Optional[int] = None,
        scale: float = 1.0,
    ):
        """Instantiate an `emulator` object.

        Parameters
        ----------
        trained_weights : `str`
            Filepath to .hdf5 file containing trained network weights, as
            saved by a `tensorflow.keras.Model.save_weights` command
        scaler : `str`
            Filepath to saved `sklearn.preprocessing.StandardScaler` object,
            fitted during network training
        input_size : `int`
            Dimensionality of input feature vector
        hidden_layer_width : `int`
            Width of hidden layers
        hidden_layer_depth : `int`
            Number of hidden layers
        activation : `func`
            Activation function to be applied to hidden layers
        batch_size: `int`
            Batch size to be used by `jax.vmap`

        Returns
        -------
        None
        """

        warn_if(
            not jax.config.read("jax_enable_x64"),
            msg="jax_enable_x64 is not enabled; this may cause numerical instability.",
        )

        # Instantiate neural network
        nn = eqx.nn.MLP(
            in_size=input_size,
            out_size=1,
            depth=hidden_layer_depth,
            width_size=hidden_layer_width,
            activation=activation,
            final_activation=final_activation,
            key=jax.random.PRNGKey(111),
        )

        # Load trained weights and biases
        weight_data = h5py.File(trained_weights, "r")

        # Load scaling parameters
        with open(scaler, "r") as f:
            _scaler = json.load(f)
            self.scaler = {
                "mean": jax.device_put(jnp.asarray(_scaler["mean"]), may_alias=True),
                "scale": jax.device_put(jnp.asarray(_scaler["scale"]), may_alias=True),
            }

        # Define helper functions with which to access MLP weights and biases
        # Needed by `eqx.tree_at`
        def get_weights(i: int) -> Callable[[eqx.nn.MLP], Array]:
            return lambda t: t.layers[i].weight

        def get_biases(i: int) -> Callable[[eqx.nn.MLP], Optional[Array]]:
            return lambda t: t.layers[i].bias

        # Loop across layers, load pre-trained weights and biases
        for i in range(hidden_layer_depth + 1):
            if i == 0:
                key = "dense"
            else:
                key = "dense_{0}".format(i)

            layer_weights = jax.device_put(
                weight_data["{0}/{0}/kernel:0".format(key)][()].T, may_alias=True
            )
            nn = eqx.tree_at(get_weights(i), nn, layer_weights)

            layer_biases = jax.device_put(
                weight_data["{0}/{0}/bias:0".format(key)][()].T, may_alias=True
            )
            nn = eqx.tree_at(get_biases(i), nn, layer_biases)

        weight_data.close()

        self.nn_vt = nn
        self.scale = scale
        self.batch_size = batch_size

    def _transform_parameters(self, *args, **kwargs) -> Array:
        """OVERWRITE UPON SUBCLASSING.

        Function to convert from a predetermined set of user-provided physical
        CBC parameters to the input space expected by the trained neural
        network. Used by `emulator.__call__` below.

        NOTE: This function should be JIT-able and differentiable, and so
        consistency/completeness checks should be performed upstream; we
        should be able to assume that `physical_params` is provided as
        expected.

        Parameters
        ----------
        *args : `jax.numpy.array`
            physical parameters characterizing CBC signals
        **kwargs : `jax.numpy.array`
            physical parameters characterizing CBC signals

        Returns
        -------
        transformed_parameters : `jax.numpy.array`
            Transformed parameter space expected by trained neural network
        """
        raise NotImplementedError

    def __call__(self, x):
        """Function to evaluate the trained neural network on a set of user- provided
        physical CBC parameters.

        NOTE: This function should be JIT-able and differentiable, and so any
        consistency or completeness checks should be performed upstream, such
        that we can assume the provided parameter vector `x` is already in the
        correct format expected by the `emulator._transform_parameters` method.
        """

        # Transform physical parameters to space expected by the neural network
        transformed_x = self._transform_parameters(
            *[x[..., i] for i in range(x.shape[-1])]
        )

        # Apply scaling, evaluate the network, and return
        scaled_x = (transformed_x - self.scaler["mean"]) / self.scaler["scale"]
        return jax.vmap(self.nn_vt)(scaled_x)

    @abstractmethod
    def check_input(
        self,
        key: PRNGKeyArray,
        shape: tuple[int, ...],
        parameter_dict: dict[str, Array],
    ) -> Tuple[PRNGKeyArray, dict[str, Array]]:
        """Method to check provided set of compact binary parameters for any missing
        information, and/or to augment provided parameters with any additional derived
        information expected by the neural network. If extrinsic parameters (e.g. sky
        location, polarization angle, etc.) have not been provided, they will be
        randomly generated and appended to the given CBC parameters.

        Parameters
        ----------
        key: `jax.random.PRNGKey`
            Random key to be used for generating extrinsic parameters
        shape: `tuple`
            Shape of the input array
        parameter_dict : `dict`
            Set of compact binary parameters for which we want to evaluate Pdet

        Returns
        -------
        parameter_dict : `dict`
            Dictionary of CBC parameters, augmented with necessary derived parameters
        """
        raise NotImplementedError



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from abc import abstractmethod
from argparse import ArgumentParser
from collections.abc import Callable
from typing import Any, Dict, List, Optional, Tuple, Union

from jax import random as jrd
from jaxtyping import Array, PRNGKeyArray
from loguru import logger
from numpyro._typing import DistributionLike
from numpyro.util import is_prng_key

from gwkokab.models.utils import JointDistribution
from gwkokab.utils.tools import error_if
from kokab.utils.common import get_processed_priors, read_json, write_json

from .bake import Bake


class Guru:
    """Guru is a class which contains all the common functionality among Genie, Sage and
    Guru classes.
    """

    _rng_key: PRNGKeyArray

    def __init__(
        self,
        *,
        analysis_name: str,
        check_leaks: bool,
        debug_nans: bool,
        model: Union[DistributionLike, Callable[..., DistributionLike]],
        poisson_mean_filename: str,
        prior_filename: str,
        profile_memory: bool,
        sampler_settings_filename: str,
        selection_fn_filename: str,
    ) -> None:
        self.analysis_name = analysis_name
        self.prior_filename = prior_filename
        self.model = model
        self.sampler_settings_filename = sampler_settings_filename
        self.debug_nans = debug_nans
        self.profile_memory = profile_memory
        self.check_leaks = check_leaks
        self.selection_fn_filename = selection_fn_filename
        self.poisson_mean_filename = poisson_mean_filename

    @property
    def rng_key(self) -> PRNGKeyArray:
        self._rng_key, subkey = jrd.split(self._rng_key)
        return subkey

    def set_rng_key(
        self, *, key: Optional[PRNGKeyArray] = None, seed: Optional[int] = None
    ) -> None:
        error_if(
            key is None and seed is None,
            msg="Either 'key' or 'seed' must be provided to set the random number generator key.",
        )
        if key is not None:
            error_if(
                not is_prng_key(key),
                msg=f"Expected a PRNGKeyArray, got {type(key)}.",
            )
            logger.info(f"Setting the random number generator key to {key}.")
            self._rng_key = key
        elif seed is not None:
            error_if(
                not isinstance(seed, int),
                msg=f"Expected an integer seed, got {type(seed)}.",
            )
            error_if(
                seed < 0,
                msg=f"Seed must be a non-negative integer, got {seed}.",
            )
            logger.info(f"Setting the random number generator key with seed {seed}.")
            key = jrd.PRNGKey(seed)
            self._rng_key = key

    def bake_model(
        self,
    ) -> Tuple[
        Dict[str, Union[int, float, bool, None]],
        Callable[..., DistributionLike],
        JointDistribution,
        Dict[str, int],
        Dict[str, int],
    ]:
        """Returns a Bake object for the model.

        Returns
        -------
        Tuple[ Dict[str, Union[int, float, bool, None]], Callable[..., DistributionLike], JointDistribution, Dict[str, int], ]
            A tuple containing the constants, the distribution function, the prior
            distribution, and the variables index.
        """
        prior_dict = read_json(self.prior_filename)
        model_prior_param = get_processed_priors(self.model_parameters, prior_dict)

        logger.debug("Baking the model")
        baked_model = Bake(self.model)(**self.constants, **model_prior_param)

        constants, variables, duplicates, dist_fn = baked_model.get_dist()  # type: ignore
        variables_index: dict[str, int] = {
            key: i for i, key in enumerate(sorted(variables.keys()))
        }
        for key, value in duplicates.items():
            variables_index[key] = variables_index[value]

        group_variables: dict[int, list[str]] = {}
        for key, value in variables_index.items():  # type: ignore
            group_variables[value] = group_variables.get(value, []) + [key]  # type: ignore

        logger.debug(
            "Number of recovering variables: {num_vars}", num_vars=len(group_variables)
        )

        for key, value in constants.items():  # type: ignore
            logger.debug(
                "Constant variable: {name} = {variable}", name=key, variable=value
            )

        for value in group_variables.values():  # type: ignore
            logger.debug("Recovering variable: {variable}", variable=", ".join(value))

        write_json("constants.json", constants)
        write_json("nf_samples_mapping.json", variables_index)

        priors = JointDistribution(
            *[variables[key] for key in sorted(variables.keys())], validate_args=True
        )

        return constants, dist_fn, priors, variables, variables_index

    @property
    def parameters(self) -> List[str]:
        """Returns the parameters (intrinsic + extrinsic).

        Returns
        -------
        List[str]
            List of parameters.

        Raises
        ------
        NotImplementedError
            If the Guru class is used directly, this method raises a NotImplementedError.
            It is expected that subclasses of Guru will implement this method.
        """
        msg = (
            "The Guru class should not be used directly. Please use a subclass that "
            "implements the parameters property."
        )
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def model_parameters(self) -> List[str]:
        """Returns the model parameters.

        Returns
        -------
        List[str]
            List of model parameters.

        Raises
        ------
        NotImplementedError
            If the Guru class is used directly, this method raises a NotImplementedError.
            It is expected that subclasses of Guru will implement this method.
        """
        msg = (
            "The Guru class should not be used directly. Please use a subclass that "
            "implements the model_parameters property."
        )
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        """Returns the constants used in the model.

        Returns
        -------
        Dict[str, Union[int, float, bool]]
            A dictionary containing the constants used in the model.
        """
        return {}

    @abstractmethod
    def driver(
        self,
        *,
        logpdf: Callable[[Array, Dict[str, Any]], Array],
        priors: JointDistribution,
        data: Dict[str, Any],
        labels: List[str],
    ) -> None:
        raise NotImplementedError()


def guru_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Guru script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    parser.add_argument(
        "--seed",
        help="Seed for the random number generator.",
        default=37,
        type=int,
    )

    pmean_group = parser.add_argument_group("Poisson Mean Options")
    pmean_group.add_argument(
        "--vt-json",
        help="Path to the JSON file containing the VT options.",
        type=str,
        default="vt.json",
    )
    pmean_group.add_argument(
        "--pmean-json",
        help="Path to the JSON file containing the Poisson mean options.",
        type=str,
        default="pmean.json",
    )

    sampler_group = parser.add_argument_group("Sampler Options")
    sampler_group.add_argument(
        "--sampler-config",
        help="Path to the JSON file containing the sampler configuration.",
        type=str,
        required=True,
    )

    prior_group = parser.add_argument_group("Prior Options")
    prior_group.add_argument(
        "--prior-json",
        type=str,
        help="Path to a JSON file containing the prior distributions.",
        default="prior.json",
    )

    debug_group = parser.add_argument_group("Debug Options")
    debug_group.add_argument(
        "--debug-nans",
        help="Checks for NaNs in each computation. See details in the documentation: "
        "https://jax.readthedocs.io/en/latest/_autosummary/jax.debug_nans.html#jax.debug_nans.",
        action="store_true",
    )
    debug_group.add_argument(
        "--profile-memory",
        help="Enable memory profiling.",
        action="store_true",
    )
    debug_group.add_argument(
        "--check-leaks",
        help="Check for JAX Tracer leaks. See details in the documentation: "
        "https://jax.readthedocs.io/en/latest/_autosummary/jax.checking_leaks.html#jax.checking_leaks.",
        action="store_true",
    )

    return parser



# Copyright (c) 2025 Tom Callister
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from .O3 import pdet_O3 as pdet_O3



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import jax
from jax import nn as jnn, numpy as jnp
from jaxtyping import ArrayLike


@jax.jit
def log_planck_taper_window(x: ArrayLike) -> ArrayLike:
    r"""If :math:`x` is the point at which to evaluate the window, then the Planck taper
    window is defined as,

    .. math::

        S(x)=\begin{cases}
            0                                                                   & \text{if } x < 0,           \\
            \displaystyle\frac{1}{1+e^{\left(\frac{1}{x}+\frac{1}{x-1}\right)}} & \text{if } 0 \leq x \leq 1, \\
            1                                                                   & \text{if } x > 1,           \\
        \end{cases}

    This function evaluates the log of the Planck taper window :math:`\ln{S(x)}`.

    Parameters
    ----------
    x: ArrayLike
        point at which to evaluate the window

    Returns
    -------
    ArrayLike
        window value
    """
    inv_1 = jnp.where(x == 0.0, 0.0, 1.0 / jnp.where(x == 0.0, 1.0, x))
    inv_2 = jnp.where(x == 1.0, 0.0, 1.0 / jnp.where(x == 1.0, 1.0, x - 1.0))
    return jnp.where(
        x <= 0,
        -jnp.inf,
        jnp.where(x < 1.0, -jnn.softplus(inv_1 + inv_2), 0.0),
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import os

import corner
import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Corner plotter.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a corner plot.",
    )
    parser.add_argument(
        "-d",
        "--data",
        help="data file path. Only .dat files are supported.",
        required=True,
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "-l",
        "--labels",
        nargs="+",
        help="labels for the corner plot",
        default=None,
    )
    parser.add_argument(
        "-t",
        "--truths",
        help="truth values for the corner plot",
        nargs="+",
        type=float,
    )
    parser.add_argument(
        "-q",
        "--quantiles",
        help="quantiles for the corner plot",
        default=None,
        type=list[float],  # type: ignore[arg-type]
    )
    parser.add_argument(
        "-b",
        "--bins",
        help="number of bins for the corner plot",
        default=20,
        type=int,
    )
    parser.add_argument(
        "-s",
        "--smooth",
        help="smooth the corner plot",
        action="store_true",
    )
    parser.add_argument(
        "-st",
        "--show-titles",
        help="show titles in the corner plot",
        action="store_true",
    )
    parser.add_argument(
        "-scale",
        "--scale",
        help="scale the corner plot",
        default=1.0,
        type=float,
    )
    parser.add_argument(
        "-size",
        "--size",
        help="size of the corner plot in inches",
        nargs=2,
        default=(6, 6),
        type=float,
    )
    parser.add_argument(
        "--range",
        help="range of the corner plot",
        nargs="+",
        action="append",
        default=None,
        type=float,
    )
    parser.add_argument(
        "--truth-color",
        help="color of the truth values in the corner plot",
        default="red",
        type=str,
    )
    parser.add_argument(
        "--color",
        help="color of the corner plot",
        default="#3498DB",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )
    parser.add_argument(
        "--title-fmt",
        help="format string for titles in the corner plot",
        default=".3f",
        type=str,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    data = pd.read_csv(args.data.name, delimiter=" ", skiprows=1).to_numpy()
    figure = corner.corner(
        data,
        labels=args.labels,
        truths=args.truths if args.truths is not None else None,
        quantiles=args.quantiles,
        bins=args.bins,
        smooth=args.smooth,
        show_titles=args.show_titles,
        truth_color=args.truth_color,
        color=args.color,
        plot_datapoints=False,
        range=args.range,
        title_fmt=args.title_fmt,
    )
    scaling_factor = args.scale
    figure.set_size_inches(scaling_factor * args.size[0], scaling_factor * args.size[1])
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Literals used in Kokab."""

from typing import Final


LOG_REF_PRIOR_NAME: Final[str] = "log_prior"
"""Helper variable to store the column name of log reference prior."""


INFERENCE_DIRECTORY: Final[str] = "inference"
"""Name of the directory to store inference results."""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import os
from collections.abc import Callable
from typing import Any, Dict, List

import arviz as az
import jax
import numpy as np
import numpyro
import pandas as pd
from jaxtyping import Array
from loguru import logger
from numpyro.infer import MCMC, NUTS

from gwkokab.models.utils import JointDistribution
from kokab.utils.common import read_json
from kokab.utils.guru import Guru, guru_arg_parser
from kokab.utils.literals import INFERENCE_DIRECTORY


def save_inference_data(mcmc: numpyro.infer.MCMC) -> None:
    os.makedirs(INFERENCE_DIRECTORY, exist_ok=True)

    inference_data = az.from_numpyro(mcmc)

    header = list(inference_data.posterior.data_vars.keys())

    posterior_samples = mcmc.get_samples()
    np.savetxt(
        INFERENCE_DIRECTORY + "/samples.dat",
        np.column_stack([posterior_samples[key] for key in header]),
        header=" ".join(header),
    )

    summary = az.summary(inference_data)

    pd.DataFrame(summary).to_json(
        INFERENCE_DIRECTORY + "/posterior_summary.json", indent=4
    )

    posterior_data = np.permute_dims(
        np.asarray(inference_data.posterior.to_dataarray()),
        (1, 2, 0),  # (variable, chain, draw) -> (chain, draw, variable)
    )

    n_chains = posterior_data.shape[0]

    for i in range(n_chains):
        np.savetxt(
            INFERENCE_DIRECTORY + f"/chain_{i}.dat",
            posterior_data[i],
            header=" ".join(header),
            comments="#",
            delimiter=" ",
        )


class NumpyroBased(Guru):
    def driver(
        self,
        *,
        logpdf: Callable[[Array, Dict[str, Any]], Array],
        priors: JointDistribution,
        data: Dict[str, Any],
        labels: List[str],
    ) -> None:
        del priors
        del labels

        sampler_config = read_json(self.sampler_settings_filename)

        def f() -> MCMC:
            kernel = NUTS(logpdf, **sampler_config["kernel"])
            mcmc = MCMC(kernel, **sampler_config["mcmc"])
            mcmc.run(self.rng_key, **data)
            return mcmc

        if self.debug_nans:
            with jax.debug_nans(True):
                mcmc = f()
        elif self.profile_memory:
            mcmc = f()

            import datetime

            time = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            filename = f"{self.analysis_name}_memory_{time}.prof"
            jax.profiler.save_device_memory_profile(filename)
        elif self.check_leaks:
            with jax.checking_leaks():
                mcmc = f()
        else:
            mcmc = f()

        save_inference_data(mcmc)

        logger.info("Sampling and data saving complete.")


numpyro_arg_parser = guru_arg_parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import os

import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Joint plot",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script creates over plots of histograms for multiple columns in a data file.",
    )
    parser.add_argument(
        "--data",
        help="data file path. Only .dat files are supported.",
        required=True,
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "--delimiter",
        help="delimiter for the data file",
        default=" ",
        type=str,
    )
    parser.add_argument(
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "--column-names",
        help="names of the columns to plot",
        required=True,
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "--labels",
        help="labels for the columns",
        default=None,
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "--xlabel",
        help="label for x-axis",
        type=str,
    )
    parser.add_argument(
        "--ylabel",
        help="label for y-axis",
        type=str,
    )
    parser.add_argument(
        "--title",
        help="title of the plot",
        type=str,
    )
    parser.add_argument(
        "--size",
        help="size of the corner plot in inches",
        nargs=2,
        default=(6, 6),
        type=float,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )
    parser.add_argument(
        "--bins",
        help="number of bins for the histogram",
        type=int,
        default=30,
    )
    parser.add_argument(
        "--alpha",
        help="transparency of the histogram bars",
        type=float,
        default=0.5,
    )
    parser.add_argument(
        "--x-scale",
        help="scale of the x-axis",
        choices=["linear", "log", "symlog"],
        type=str,
        default="linear",
    )
    parser.add_argument(
        "--y-scale",
        help="scale of the y-axis",
        choices=["linear", "log", "symlog"],
        type=str,
        default="linear",
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()
    if args.labels is None:
        labels = args.column_names
    else:
        assert len(args.labels) == len(args.column_names), (
            "Number of labels must match number of column names."
        )
        labels = args.labels

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    with open(args.data.name) as f:
        header = f.readline().lstrip("#").strip().split()

    data = pd.read_csv(args.data.name, sep=r"\s+", comment="#", header=None)
    data.columns = header

    plt.figure(figsize=args.size)
    for x_column in args.column_names:
        if x_column not in data.columns:
            raise ValueError(f"Column '{x_column}' not found in data.")

    for x_column, label in zip(args.column_names, labels):
        plt.hist(
            data[x_column],
            bins=args.bins,
            density=True,
            alpha=args.alpha,
            label=label,
        )
    if args.xlabel is not None:
        plt.xlabel(args.xlabel)
    if args.ylabel is not None:
        plt.ylabel(args.ylabel)
    if args.title is not None:
        plt.title(args.title)
    plt.legend()
    plt.xscale(args.x_scale)
    plt.yscale(args.y_scale)
    plt.tight_layout()

    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Dict, List, Literal, Optional, Tuple, Union

import equinox as eqx
import jax
from jax import nn as jnn, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from loguru import logger
from numpyro._typing import DistributionLike
from numpyro.distributions.distribution import Distribution

from .models import PowerlawRedshift
from .models.utils import JointDistribution, ScaledMixture
from .utils.tools import batch_and_remainder, error_if
from .vts import (
    SemiAnalyticalRealInjectionVolumeTimeSensitivity,
    SyntheticInjectionVolumeTimeSensitivity,
    VolumeTimeSensitivityInterface,
)


class PoissonMean(eqx.Module):
    r"""
    Inverse Transform Sampling
    --------------------------

    Samples are generated from :math:`\rho_{\Omega\mid\Lambda}` by using the inverse
    transform sampling method. The estimator is given by,

    .. math::

        \hat{\mu}_{\Omega\mid\Lambda} \approx \frac{1}{N}\sum_{i=1}^{N}\operatorname{VT}(\omega_i),
        \qquad \forall 0<i\leq N, \omega_i \sim \rho_{\Omega\mid\Lambda}.

    This method is very useful when the target distribution is easy to sample from.

    Importance Sampling
    -------------------

    It is very cheap to generate random samples from the proposal distribution and
    evaluate the importance weights and reusing them for rest of the calculations. We
    assume a proposal distribution :math:`\rho_{\phi}` and calculate the importance
    weights as,

    .. math::
        w_i = \frac{\operatorname{VT(\omega_i)}}{\rho_{\phi}(\omega_i)},
        \qquad \forall 0<i\leq N, \omega_i \sim \rho_{\phi}.

    We save these weights and samples for the later use and estimate
    :math:`\mu_{\Omega\mid\Lambda}` as,

    .. math::
        \hat{\mu}_{\Omega\mid\Lambda} \approx \frac{1}{N} \sum_{i=1}^{N}
        w_i \cdot \rho_{\Omega\mid\Lambda}(\omega_i\mid\lambda).

    Importance sampling is a variance reduction technique and it is very useful when the
    proposal distribution is close to the target distribution. Choice of a good proposal
    distribution is a tedious task and it is not always possible to find a good proposal
    distribution. To overcome this problem, we have used a mixture of proposal
    distributions to sample from. Usually, we use a uniform distribution and a
    log-uniform distribution over an appropriate range of the parameters. Some times we
    also add a peak at the maxima and at the expectation of the proposal distribution to
    improve the performance of the importance sampling.
    """

    is_injection_based: bool = eqx.field(default=False, static=True)
    """Flag to check if the class is injection based or not."""
    key: PRNGKeyArray
    logVT_estimator: Optional[VolumeTimeSensitivityInterface] = eqx.field(default=None)
    num_samples_per_component: Optional[List[int]] = eqx.field(
        static=True, default=None
    )
    proposal_log_weights_and_samples: Tuple[Optional[Tuple[Array, Array]], ...]
    time_scale: Union[int, float, Array] = eqx.field(default=1.0)
    parameter_ranges: Optional[Dict[str, Union[int, float]]] = eqx.field(default=None)

    def __init__(
        self,
        logVT_estimator: VolumeTimeSensitivityInterface,
        key: PRNGKeyArray,
        proposal_dists: Optional[List[Union[Literal["self"], DistributionLike]]] = None,
        num_samples: int = 10_000,
        self_num_samples: Optional[int] = None,
        num_samples_per_component: Optional[List[int]] = None,
        time_scale: Union[int, float, Array] = 1.0,
    ) -> None:
        """
        Parameters
        ----------
        logVT_estimator : VolumeTimeSensitivityInterface
            Volume Time Sensitivity estimator.
        key : PRNGKeyArray
            PRNG key.
        proposal_dists : List[Union[Literal[&quot;self&quot;], DistributionLike]]
            List of proposal distributions. If "self" is given, the proposal distribution
            is the target distribution itself.
        num_samples : int
            Number of samples, by default 10_000
        self_num_samples : Optional[int], optional
            Number of samples for distribution using Inverse Transform Sampling, by default None
        num_samples_per_component : Optional[List[int]], optional
            Number of samples for each component, by default None
        time_scale : Union[int, float, Array]
            scale factor for time, by default 1.0

        Raises
        ------
        ValueError
            If the proposal distribution is unknown.
        ValueError
            If the proposal distribution is not a distribution.
        """
        if hasattr(logVT_estimator, "parameter_ranges"):
            self.parameter_ranges = logVT_estimator.parameter_ranges
        if isinstance(
            logVT_estimator,
            (
                SyntheticInjectionVolumeTimeSensitivity,
                SemiAnalyticalRealInjectionVolumeTimeSensitivity,
            ),
        ):
            self.is_injection_based = True
            self.key = key
            self.proposal_log_weights_and_samples = (
                (jnp.log(logVT_estimator.sampling_prob), logVT_estimator.injections),
            )
            logger.warning(
                "The time scale is not used for injection based VTs. "
                "We parse the injection files to get the time scales."
            )
            self.time_scale = logVT_estimator.analysis_time_years
            self.num_samples_per_component = [
                logVT_estimator.total_injections,
                logVT_estimator.batch_size,
            ]

        else:
            self.logVT_estimator = logVT_estimator
            self.__init_for_per_component_rate__(
                key=key,
                proposal_dists=proposal_dists,
                num_samples=num_samples,
                self_num_samples=self_num_samples,
                num_samples_per_component=num_samples_per_component,
                time_scale=time_scale,
            )

    def __init_for_per_component_rate__(
        self,
        key: PRNGKeyArray,
        proposal_dists: List[Union[Literal["self"], DistributionLike]],
        num_samples: int = 10_000,
        self_num_samples: Optional[int] = None,
        num_samples_per_component: Optional[List[int]] = None,
        time_scale: Union[int, float, Array] = 1.0,
    ) -> None:
        """
        Parameters
        ----------
        key : PRNGKeyArray
            PRNG key.
        proposal_dists : List[Union[Literal[&quot;self&quot;], DistributionLike]]
            List of proposal distributions. If "self" is given, the proposal distribution
            is the target distribution itself.
        num_samples : int
            Number of samples, by default 10_000
        self_num_samples : Optional[int], optional
            Number of samples for distribution using Inverse Transform Sampling, by default None
        num_samples_per_component : Optional[List[int]], optional
            Number of samples for each component, by default None
        time_scale : Union[int, float, Array]
            scale factor for time, by default 1.0

        Raises
        ------
        ValueError
            If the proposal distribution is unknown.
        ValueError
            If the proposal distribution is not a distribution.
        """
        if num_samples_per_component is not None:
            error_if(
                len(proposal_dists) != len(num_samples_per_component),
                AssertionError,
                "Mismatch between the number of proposal distributions "
                f"({len(proposal_dists)}) and the number of samples per component "
                f"({len(num_samples_per_component)})",
            )
            self.num_samples_per_component = num_samples_per_component
        else:
            self.num_samples_per_component = [
                num_samples for _ in range(len(proposal_dists))
            ]
        self.time_scale = time_scale
        logVT_fn = self.logVT_estimator.get_mapped_logVT()  # type: ignore

        proposal_log_weights_and_samples = []
        for index, dist in enumerate(proposal_dists):
            if isinstance(dist, str):
                if dist.strip().lower() == "self":
                    proposal_log_weights_and_samples.append(None)
                    if self_num_samples is not None:
                        self.num_samples_per_component[index] = self_num_samples
                else:
                    raise ValueError(f"Unknown proposal distribution: {dist}")
            elif isinstance(dist, Distribution):
                _num_samples = self.num_samples_per_component[index]
                samples = dist.sample(key, (_num_samples,))
                proposal_log_prob: Array = dist.log_prob(samples).reshape(_num_samples)
                logVT_samples = logVT_fn(samples).reshape(_num_samples)
                log_weights = logVT_samples - proposal_log_prob
                error_if(
                    log_weights.shape != (_num_samples,),
                    AssertionError,
                    f"Expected log_weights to have shape {(_num_samples,)}, "
                    f"but got {log_weights.shape}",
                )
                error_if(
                    samples.shape[0] != _num_samples,
                    AssertionError,
                    f"Expected samples to have shape {(_num_samples, -1)}, "
                    f"but got {samples.shape}",
                )
                error_if(
                    not jnp.all(jnp.isfinite(log_weights)),
                    AssertionError,
                    f"Expected log_weights to be finite, but got {log_weights}",
                )
                error_if(
                    not jnp.all(jnp.isfinite(samples)),
                    AssertionError,
                    f"Expected samples to be finite, but got {samples}",
                )
                proposal_log_weights_and_samples.append((log_weights, samples))
                key, _ = jrd.split(key)
            else:
                raise ValueError(f"Unknown proposal distribution: {dist}")
        self.proposal_log_weights_and_samples = tuple(proposal_log_weights_and_samples)
        self.key = key

    def __call__(self, model_instance: ScaledMixture) -> Array:
        r"""Estimate the rate/s by using the given model.

        Parameters
        ----------
        model : ScaledMixture
            Model instance.

        Returns
        -------
        Array
            Estimated rate/s.
        """

        if not self.is_injection_based:  # per component rate estimation
            return self.calculate_per_component_rate(model_instance)

        # injection based sampling method

        log_prob_fn = eqx.filter_jit(eqx.filter_vmap(model_instance.log_prob))

        def _f(carry_logsumexp: Array, data: Tuple[Array, Array]) -> Tuple[Array, None]:
            log_weights, samples = data
            # log p(θ_i|λ)
            model_log_prob = log_prob_fn(samples).reshape(log_weights.shape[0])
            # log p(θ_i|λ) - log w_i
            log_prob = model_log_prob - log_weights
            safe_log_prob = jnp.where(
                jnp.isneginf(log_prob) | jnp.isnan(log_prob),
                -jnp.inf,
                log_prob,
            )

            partial_logsumexp = jnn.logsumexp(
                safe_log_prob,
                where=~jnp.isneginf(safe_log_prob),
                axis=-1,
            )
            safe_carry_logsumexp = jnp.where(
                jnp.isneginf(carry_logsumexp) | jnp.isnan(carry_logsumexp),
                -jnp.inf,
                carry_logsumexp,
            )
            return jnp.logaddexp(safe_carry_logsumexp, partial_logsumexp), None

        # n_total = n_accepted + n_rejected, batch size
        num_samples, batch_size = self.num_samples_per_component  # type: ignore

        # log w_i, θ_i
        log_weights, samples = self.proposal_log_weights_and_samples[0]  # type: ignore

        n_accepted = log_weights.shape[0]
        initial_logprob = jnp.asarray(-jnp.inf)
        if n_accepted <= batch_size:
            # If the number of accepted injections is less than or equal to the batch size,
            # we can process them all at once.
            log_prob, _ = _f(initial_logprob, (log_weights, samples))
        else:
            batched_log_weights, remainder_log_weights = batch_and_remainder(
                log_weights, batch_size
            )
            batched_samples, remainder_samples = batch_and_remainder(
                samples, batch_size
            )
            batched_logprob, _ = jax.lax.scan(
                _f,
                initial_logprob,
                (batched_log_weights, batched_samples),
            )
            log_prob, _ = _f(
                batched_logprob, (remainder_log_weights, remainder_samples)
            )

        # (T / n_total) * exp(log Σ exp(log p(θ_i|λ) - log w_i))
        return (self.time_scale / num_samples) * jnp.exp(log_prob)

    def calculate_per_component_rate(self, model: ScaledMixture) -> Array:
        r"""Estimate the per component rate/s by using the given model.

        Parameters
        ----------
        model : ScaledMixture
            Model instance.

        Returns
        -------
        Array
            Estimated rate/s.
        """
        error_if(
            not isinstance(model, ScaledMixture),
            msg=f"Expected model to be an instance of ScaledMixture, but got {type(model)}",
        )
        error_if(
            len(self.proposal_log_weights_and_samples) != model.mixture_size,
            msg=f"Expected {model.mixture_size} proposal distributions, "
            f"but got {len(self.proposal_log_weights_and_samples)}",
        )

        log_estimated_rate = jnp.asarray(-jnp.inf)

        logVT_fn = self.logVT_estimator.get_mapped_logVT()  # type: ignore

        for i in range(model.mixture_size):
            log_weights_and_samples = self.proposal_log_weights_and_samples[i]
            component_dist: DistributionLike = model.component_distributions[i]
            num_samples = self.num_samples_per_component[i]  # type: ignore
            if (
                log_weights_and_samples is None
            ):  # case 1: "self" meaning inverse transform sampling
                samples = jax.jit(component_dist.sample, static_argnums=(1,))(
                    self.key, (num_samples,)
                )
                per_sample_log_estimated_rates = logVT_fn(samples)
                if isinstance(component_dist, JointDistribution):
                    for m_dist in component_dist.marginal_distributions:
                        if isinstance(m_dist, PowerlawRedshift):
                            per_sample_log_estimated_rates += m_dist.log_norm()
                            break
            else:  # case 2: importance sampling
                log_weights, samples = log_weights_and_samples
                component_log_prob = jax.checkpoint(jax.jit(component_dist.log_prob))(
                    samples
                ).reshape(num_samples)
                per_sample_log_estimated_rates = log_weights + component_log_prob

            log_rate_i = jax.lax.dynamic_index_in_dim(
                model.log_scales, index=i, axis=-1, keepdims=False
            )
            log_estimated_rate_i = (
                log_rate_i
                - jnp.log(num_samples)
                + jnn.logsumexp(
                    per_sample_log_estimated_rates,
                    where=~jnp.isneginf(per_sample_log_estimated_rates),
                    axis=-1,
                )
            )

            log_estimated_rate = jnp.logaddexp(log_estimated_rate, log_estimated_rate_i)

        return self.time_scale * jnp.exp(log_estimated_rate)
